# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This module contains FreeCAD commands for the BIM workbench"""

import os
import sys

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class BIM_IfcProperties:

    def GetResources(self):
        return {
            "Pixmap": "BIM_IfcProperties",
            "MenuText": QT_TRANSLATE_NOOP("BIM_IfcProperties", "Manage IFC Properties"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_IfcProperties",
                "Manages the different IFC properties of the BIM objects",
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        # only raise the dialog if it is already open
        if getattr(self, "form", None):
            self.form.raise_()
            return

        from PySide import QtGui

        try:
            import ArchIFC

            self.ifcroles = ArchIFC.IfcTypes
        except (ImportError, AttributeError):
            import ArchComponent

            self.ifcroles = ArchComponent.IfcRoles

        # load the form and set the tree model up
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/dialogIfcProperties.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/BIM_IfcProperties.svg"))
        self.form.groupMode.setItemIcon(2, QtGui.QIcon(":/icons/Document.svg"))
        self.model = QtGui.QStandardItemModel()
        self.form.tree.setModel(self.model)
        self.form.tree.setUniformRowHeights(True)

        # restore saved values
        self.form.onlySelected.setChecked(PARAMS.GetInt("IfcPropertiesSelectedState", 0))
        self.form.onlyVisible.setChecked(PARAMS.GetInt("IfcPropertiesVisibleState", 0))
        w = PARAMS.GetInt("BimIfcPropertiesDialogWidth", 1200)
        h = PARAMS.GetInt("BimIfcPropertiesDialogHeight", 608)
        self.form.resize(w, h)

        # build objects list and fill search terms
        self.objectslist, searchterms = self.rebuildObjectsList()
        self.form.searchField.addItems(searchterms)

        # set the properties editor
        try:
            import ArchIFCSchema

            self.ptypes = list(ArchIFCSchema.IfcTypes.keys())
        except (ImportError, AttributeError):
            import ArchComponent

            self.ptypes = ArchComponent.SimplePropertyTypes + ArchComponent.MeasurePropertyTypes
        self.plabels = [
            "".join(map(lambda x: x if x.islower() else " " + x, t[3:]))[1:] for t in self.ptypes
        ]
        self.psetdefs = {}
        psetpath = os.path.join(
            FreeCAD.getResourceDir(), "Mod", "BIM", "Presets", "pset_definitions.csv"
        )
        custompath = os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "CustomPsets.csv")
        self.psetdefs = self.readFromCSV(psetpath)
        self.psetdefs.update(self.readFromCSV(custompath))
        self.psetkeys = [
            "".join(map(lambda x: x if x.islower() else " " + x, t[5:]))[1:]
            for t in self.psetdefs.keys()
        ]
        self.psetkeys.sort()
        self.propmodel = QtGui.QStandardItemModel()
        self.form.treeProperties.setModel(self.propmodel)
        # self.ifcEditor.treeProperties.setDragDropMode(QtGui.QAbstractItemView.InternalMove)
        self.form.treeProperties.setUniformRowHeights(True)
        self.form.treeProperties.setItemDelegate(
            propertiesDelegate(container=self, ptypes=self.ptypes, plabels=self.plabels)
        )
        self.form.labelinfo.setText(
            self.form.labelinfo.text()
            + " "
            + translate("BIM", "Custom property sets can be defined in")
            + " "
            + custompath
        )

        # set combos
        self.form.comboProperty.addItems([translate("BIM", "Add property")] + self.plabels)
        self.form.comboPset.addItems(
            [translate("BIM", "Add property set"), translate("BIM", "New")] + self.psetkeys
        )

        # connect signals
        self.form.tree.selectionModel().selectionChanged.connect(self.updateProperties)
        self.form.groupMode.currentIndexChanged.connect(self.update)
        if hasattr(self.form.onlyVisible, "checkStateChanged"):  # Qt version >= 6.7.0
            self.form.onlyVisible.checkStateChanged.connect(self.update)
            self.form.onlySelected.checkStateChanged.connect(self.onSelected)
            self.form.onlyMatches.checkStateChanged.connect(self.update)
        else:  # Qt version < 6.7.0
            self.form.onlyVisible.stateChanged.connect(self.update)
            self.form.onlySelected.stateChanged.connect(self.onSelected)
            self.form.onlyMatches.stateChanged.connect(self.update)
        self.form.buttonBox.accepted.connect(self.accept)
        self.form.rejected.connect(self.reject)  # also triggered by self.form.buttonBox.rejected
        self.form.searchField.currentIndexChanged.connect(self.update)
        self.form.searchField.editTextChanged.connect(self.update)
        self.form.comboProperty.currentIndexChanged.connect(self.addProperty)
        self.form.comboPset.currentIndexChanged.connect(self.addPset)
        self.form.buttonIFCPropertiesDelete.clicked.connect(self.removeProperty)
        self.form.treeProperties.setSortingEnabled(True)

        # center the dialog over FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.form.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.form.rect().center()
        )

        self.update()
        self.form.show()

    def rebuildObjectsList(self):

        # build objects list and fill search terms
        objectslist = {}
        searchterms = [""]
        if self.form.onlySelected.isChecked():
            objects = FreeCADGui.Selection.getSelection()
        else:
            objects = FreeCAD.ActiveDocument.Objects
        for obj in objects:
            role = self.getRole(obj)
            if role:
                if hasattr(obj, "IfcProperties") and isinstance(obj.IfcProperties, dict):
                    props = obj.IfcProperties
                elif hasattr(obj, "IfcClass"):
                    props = self.getNativeIfcProperties(obj)
                else:
                    props = {}
                objectslist[obj.Name] = [role, props]
                for key, val in props.items():
                    val = val.split(";;")
                    if ";;" in key:
                        # 0.19 format
                        # pset;;pname = ptype;;pvalue
                        key = key.split(";;")
                        val = [key[1]] + val
                        key = key[0]
                    if not key in searchterms:
                        searchterms.append(key)
                    if len(val) == 3:
                        if not val[0] in searchterms:
                            searchterms.append(val[0])
        return objectslist, searchterms

    def update(self, index=None):
        "updates the tree widgets in all tabs"

        self.model.clear()
        self.model.setHorizontalHeaderLabels(
            [
                translate("BIM", "Label"),
                translate("BIM", "IFC type"),
                translate("BIM", "Search results"),
            ]
        )
        # self.form.tree.header().setResizeMode(QtGui.QHeaderView.Stretch)
        # self.form.tree.resizeColumnsToContents()
        if self.form.groupMode.currentIndex() == 1:
            # group by type
            self.updateByType()
        elif self.form.groupMode.currentIndex() == 2:
            # group by model structure
            self.updateByTree()
        else:
            # group alphabetically
            self.updateDefault()
        self.model.sort(0)
        self.form.tree.setColumnWidth(0, 300)

    def getRole(self, obj):
        if hasattr(obj, "IfcType"):
            return obj.IfcType
        elif hasattr(obj, "IfcRole"):
            return obj.IfcRole
        elif hasattr(obj, "IfcClass"):
            return obj.IfcClass
        else:
            return None

    def readFromCSV(self, csvfile):
        """reads a csv file and returns a dict"""

        import csv

        result = {}
        if os.path.exists(csvfile):
            with open(csvfile, "r") as f:
                reader = csv.reader(f, delimiter=";")
                for row in reader:
                    result[row[0]] = row[1:]
        return result

    def updateByType(self):
        from PySide import QtGui

        groups = {}
        for name, role in self.objectslist.items():
            role = role[0]
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                    groups.setdefault(role, []).append(name)

        for group in groups.keys():
            s1 = group + " (" + str(len(groups[group])) + ")"
            top = QtGui.QStandardItem(s1)
            self.model.appendRow([top, QtGui.QStandardItem(), QtGui.QStandardItem()])
            for name in groups[group]:
                obj = FreeCAD.ActiveDocument.getObject(name)
                if obj:
                    it1 = QtGui.QStandardItem(obj.Label)
                    icon = obj.ViewObject.Icon
                    it1.setIcon(icon)
                    it1.setToolTip(obj.Name)
                    it2 = QtGui.QStandardItem(group)
                    if group != self.getRole(obj):
                        it2.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                    it3 = self.getSearchResults(obj)
                    if it3:
                        top.appendRow([it1, it2, it3])
            top.sortChildren(0)
        self.form.tree.expandAll()
        self.spanTopLevels()

    def updateByTree(self):
        from PySide import QtGui

        # order by hierarchy
        def istop(obj):
            for parent in obj.InListRecursive:
                if parent.Name in self.objectslist.keys():
                    return False
            return True

        rel = []
        deps = []
        for name in self.objectslist.keys():
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if istop(obj):
                    rel.append(obj)
                else:
                    deps.append(obj)
        pa = 1
        while deps:
            for obj in rel:
                for child in obj.OutList:
                    if child in deps:
                        rel.append(child)
                        deps.remove(child)
            pa += 1
            if pa == 10:  # max 10 hierarchy levels, okay? Let's keep civilised
                rel.extend(deps)
                break

        done = {}
        for obj in rel:
            role = self.objectslist[obj.Name][0]
            if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                it1 = QtGui.QStandardItem(obj.Label)
                icon = obj.ViewObject.Icon
                it1.setIcon(icon)
                it1.setToolTip(obj.Name)
                it2 = QtGui.QStandardItem(role)
                if role != self.getRole(obj):
                    it2.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                it3 = self.getSearchResults(obj)
                ok = False
                for par in obj.InListRecursive:
                    if par.Name in done:
                        if (not hasattr(par, "Hosts")) or (obj not in par.Hosts):
                            if it3:
                                done[par.Name].appendRow([it1, it2, it3])
                            done[obj.Name] = it1
                            ok = True
                            break
                if not ok:
                    if it3:
                        self.model.appendRow([it1, it2, it3])
                    done[obj.Name] = it1
        self.form.tree.expandAll()

    def updateDefault(self):
        from PySide import QtGui

        for name, role in self.objectslist.items():
            role = role[0]
            obj = FreeCAD.ActiveDocument.getObject(name)
            if obj:
                if (not self.form.onlyVisible.isChecked()) or obj.ViewObject.isVisible():
                    it1 = QtGui.QStandardItem(obj.Label)
                    icon = obj.ViewObject.Icon
                    it1.setIcon(icon)
                    it1.setToolTip(obj.Name)
                    it2 = QtGui.QStandardItem(role)
                    if role != self.getRole(obj):
                        it2.setIcon(QtGui.QIcon(":/icons/edit-edit.svg"))
                    it3 = self.getSearchResults(obj)
                    if it3:
                        self.model.appendRow([it1, it2, it3])

    def spanTopLevels(self):
        if self.form.groupMode.currentIndex() in [1, 2]:
            idx = self.model.invisibleRootItem().index()
            for i in range(self.model.rowCount()):
                if self.model.item(i, 0).hasChildren():
                    self.form.tree.setFirstColumnSpanned(i, idx, True)

    def accept(self):
        PARAMS.SetInt("BimIfcPropertiesDialogWidth", self.form.width())
        PARAMS.SetInt("BimIfcPropertiesDialogHeight", self.form.height())
        self.form.hide()

        # print(self.objectslist)
        changed = False
        for key, values in self.objectslist.items():
            obj = FreeCAD.ActiveDocument.getObject(key)
            if obj:
                if hasattr(obj, "IfcProperties"):
                    if not isinstance(obj.IfcProperties, dict):
                        FreeCAD.Console.PrintWarning(
                            translate(
                                "BIM",
                                "Warning: object %1 has old-styled IfcProperties and cannot be updated",
                            ).replace("%1", obj.Label)
                            + "\n"
                        )
                        continue
                    props = obj.IfcProperties
                elif hasattr(obj, "IfcClass"):
                    props = self.getNativeIfcProperties(obj)
                else:
                    props = {}
                if values[1] != props:
                    if not changed:
                        FreeCAD.ActiveDocument.openTransaction("Change properties")
                        changed = True
                    if hasattr(obj, "IfcClass"):
                        print("props:", props)
                        for key, value in values[1].items():
                            if ";;" in key and ";;" in value:
                                pname, pset = key.split(";;")
                                ptype, pvalue = value.split(";;")
                                from nativeifc import ifc_psets  # lazy loading

                                fctype = ifc_psets.get_freecad_type(ptype)
                                if not pname in obj.PropertiesList:
                                    obj.addProperty(fctype, pname, pset, ptype + ":" + pname)
                                    ifc_psets.edit_pset(obj, pname, force=True)
                                if pvalue:
                                    setattr(obj, pname, pvalue)
                    elif not hasattr(obj, "IfcProperties"):
                        obj.addProperty(
                            "App::PropertyMap",
                            "IfcPRoperties",
                            "IFC",
                            QT_TRANSLATE_NOOP("App::Property", "IFC properties of this object"),
                            locked=True,
                        )
                    if hasattr(obj, "IfcProperties"):
                        obj.IfcProperties = values[1]
        if changed:
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()
        return self.reject()

    def reject(self):
        self.form.hide()
        del self.form
        return True

    def getNativeIfcProperties(self, obj):
        props = {}
        for p in obj.PropertiesList:
            pset = obj.getGroupOfProperty(p)
            ttip = obj.getDocumentationOfProperty(p)
            if ":" in ttip:
                ptype, pname = ttip.split(":")
                if pset not in ["Base", "IFC", "Geometry"]:
                    props[pname + ";;" + pset] = ptype + ";;" + str(getattr(obj, p))
        return props

    def getSearchResults(self, obj):
        from PySide import QtGui

        text = self.form.searchField.currentText()
        if not text:
            return QtGui.QStandardItem()
        else:
            if obj.Name in self.objectslist:
                result = []
                for key, value in self.objectslist[obj.Name][1].items():
                    if ";;" in key:
                        # 0.19 format
                        key = key.split(";;")
                        pset = key[1]
                        key = key[0]
                    else:
                        pset = value.split(";;")[0]
                    if text.lower() in key.lower():
                        if not key in result:
                            result.append(key)
                    if text.lower() in pset.lower():
                        if not pset in result:
                            result.append(pset)
                if result:
                    return QtGui.QStandardItem(",".join(result))
                else:
                    if self.form.onlyMatches.isChecked():
                        return None
                    else:
                        return QtGui.QStandardItem()
            else:
                return QtGui.QStandardItem()

    def updateProperties(self, sel1=None, sel2=None):
        from PySide import QtGui

        self.propmodel.clear()
        self.propmodel.setHorizontalHeaderLabels(
            [
                translate("Arch", "Property", None),
                translate("Arch", "Type", None),
                translate("Arch", "Value", None),
            ]
        )

        self.form.treeProperties.setColumnWidth(0, 300)
        self.form.treeProperties.setColumnWidth(1, 200)

        # gather common properties
        allprops = []
        sel = self.form.tree.selectedIndexes()
        for index in sel:
            if index.column() == 0:
                name = self.model.itemFromIndex(index).toolTip()
                if name in self.objectslist:
                    allprops.append(self.objectslist[name][1])
        psets = {}
        if allprops:
            for key in allprops[0].keys():
                value = allprops[0][key].split(";;")
                iscommon = True
                if len(value) == 3:
                    pset = value[0]
                    ptype = value[1]
                    pvalue = value[2]
                elif (len(value) == 2) and (";;" in key):  # 0.19 format
                    pset = key.split(";;")[1]
                    ptype = value[0]
                    pvalue = value[1]
                elif len(value) == 2:  # old system
                    pset = "Default property set"
                    ptype = value[0]
                    pvalue = value[1]
                else:
                    print("Error: Unparsable property:", value)
                    return
                for other in allprops[1:]:
                    if key in other.keys():
                        othervalue = other[key].split(";;")
                        if len(value) == 3:
                            otherpset = othervalue[0]
                            otherptype = othervalue[1]
                            otherpvalue = othervalue[2]
                        elif (len(value) == 2) and (";;" in key):  # 0.19 format
                            otherpset = key.split(";;")[1]
                            otherptype = othervalue[0]
                            otherpvalue = othervalue[1]
                        elif len(value) == 2:  # old system
                            otherpset = "Default property set"
                            otherptype = othervalue[0]
                            otherpvalue = othervalue[1]
                        else:
                            print("Error: Unparsable property:", othervalue)
                            return
                        if otherpset != pset:
                            iscommon = False
                        if otherptype != ptype:
                            iscommon = False
                        if otherpvalue != pvalue:
                            pvalue = "*VARIES*"
                    else:
                        iscommon = False
                if iscommon:
                    plabel = ptype
                    if ptype in self.ptypes:
                        plabel = self.plabels[self.ptypes.index(ptype)]
                    psets.setdefault(pset, []).append([key, plabel, pvalue])

        # fill the tree
        for pset, plists in psets.items():
            top = QtGui.QStandardItem(pset)
            top.setDragEnabled(False)
            top.setToolTip("PropertySet")
            self.propmodel.appendRow([top, QtGui.QStandardItem(), QtGui.QStandardItem()])
            for plist in plists:
                pname = plist[0]
                if ";;" in pname:
                    pname = pname.split(";;")[0]
                it1 = QtGui.QStandardItem(pname)
                it1.setDropEnabled(False)
                it2 = QtGui.QStandardItem(plist[1])
                it2.setDropEnabled(False)
                it3 = QtGui.QStandardItem(plist[2])
                it3.setDropEnabled(False)
                top.appendRow([it1, it2, it3])
            top.sortChildren(0)
        # span top levels
        idx = self.propmodel.invisibleRootItem().index()
        for i in range(self.propmodel.rowCount()):
            if self.propmodel.item(i, 0).hasChildren():
                self.form.treeProperties.setFirstColumnSpanned(i, idx, True)
        self.form.treeProperties.expandAll()

    def updateDicts(self, remove=None):
        # update the stored dicts to reflect the editor

        sel = self.form.tree.selectedIndexes()

        for row in range(self.propmodel.rowCount()):
            pset = self.propmodel.item(row, 0).text()
            if self.propmodel.item(row, 0).hasChildren():
                for childrow in range(self.propmodel.item(row, 0).rowCount()):
                    prop = self.propmodel.item(row, 0).child(childrow, 0).text()
                    if ";;" in prop:
                        prop = prop.split(";;")[0]
                    ptype = self.propmodel.item(row, 0).child(childrow, 1).text()
                    if not ptype.startswith("Ifc"):
                        ptype = self.ptypes[self.plabels.index(ptype)]
                    pvalue = self.propmodel.item(row, 0).child(childrow, 2).text()
                    if (sys.version_info.major < 3) and isinstance(prop, unicode):
                        prop = prop.encode("utf8")

                    # update objects
                    for index in sel:
                        if index.column() == 0:
                            name = self.model.itemFromIndex(index).toolTip()
                            if name in self.objectslist:
                                # print("object",name,self.objectslist[name][1])
                                if pvalue == "*VARIES*":
                                    if not (prop + ";;" + pset in self.objectslist[name][1]):
                                        # print("adding",prop)
                                        self.objectslist[name][1][prop + ";;" + pset] = ptype + ";;"
                                else:
                                    pval = ptype + ";;" + pvalue
                                    if prop in self.objectslist[name][1]:
                                        if self.objectslist[name][1][prop + ";;" + pset] != pval:
                                            # print("modifying",prop)
                                            self.objectslist[name][1][prop + ";;" + pset] = pval
                                    else:
                                        # print("adding",prop)
                                        self.objectslist[name][1][prop + ";;" + pset] = pval
        if remove:
            for index in sel:
                if index.column() == 0:
                    name = self.model.itemFromIndex(index).toolTip()
                    if name in self.objectslist:
                        for prop in remove:
                            if prop in self.objectslist[name][1]:
                                # print("deleting",prop)
                                del self.objectslist[name][1][prop]

    def addProperty(self, idx=0, pset=None, prop=None, ptype=None):
        from PySide import QtGui

        if not self.form.tree.selectedIndexes():
            return
        if not pset:
            sel = self.form.treeProperties.selectedIndexes()
            if sel:
                item = self.propmodel.itemFromIndex(sel[0])
                if item.toolTip() == "PropertySet":
                    pset = item
        if pset:
            if not prop:
                basename = translate("Arch", "New property", None)
                # check for duplicate name
                names = []
                for i in range(self.propmodel.rowCount()):
                    topitem = self.propmodel.item(i, 0)
                    names.append(topitem.text())
                    if topitem.hasChildren():
                        for j in range(topitem.rowCount()):
                            childitem = topitem.child(j, 0)
                            names.append(childitem.text())
                suffix = 1
                newname = basename
                while newname in names:
                    newname = basename + str(suffix).zfill(3)
                    suffix += 1
                prop = newname
            if not ptype:
                if idx > 0:
                    ptype = self.plabels[idx - 1]
            if prop and ptype:
                if ptype in self.ptypes:
                    ptype = self.plabels[self.ptypes.index(ptype)]
                it1 = QtGui.QStandardItem(prop)
                it1.setDropEnabled(False)
                it2 = QtGui.QStandardItem(ptype)
                it2.setDropEnabled(False)
                it3 = QtGui.QStandardItem()
                it3.setDropEnabled(False)
                pset.appendRow([it1, it2, it3])
                self.updateDicts()
        else:
            if idx != 0:
                QtGui.QMessageBox.critical(
                    None,
                    "Error",
                    translate(
                        "BIM",
                        "Please select or create a property set first in which the new property should be placed.",
                    ),
                    QtGui.QMessageBox.Ok,
                )
        if idx != 0:
            self.form.comboProperty.setCurrentIndex(0)

    def addPset(self, idx):
        from PySide import QtGui

        if not self.form.tree.selectedIndexes():
            return
        if idx == 1:
            name = translate("Arch", "New property set", None)
            res = QtGui.QInputDialog.getText(
                None,
                translate("BIM", "New property set"),
                translate("BIM", "Property set name:"),
                QtGui.QLineEdit.Normal,
                name,
            )
            if res[1]:
                name = res[0]
            top = QtGui.QStandardItem(name)
            top.setDragEnabled(False)
            top.setToolTip("PropertySet")
            self.propmodel.appendRow([top, QtGui.QStandardItem(), QtGui.QStandardItem()])
        elif idx > 1:
            psetlabel = self.psetkeys[idx - 2]
            psetdef = "Pset_" + psetlabel.replace(" ", "")
            if psetdef in self.psetdefs:
                top = QtGui.QStandardItem(psetdef)
                top.setDragEnabled(False)
                top.setToolTip("PropertySet")
                self.propmodel.appendRow([top, QtGui.QStandardItem(), QtGui.QStandardItem()])
                for i in range(0, len(self.psetdefs[psetdef]), 2):
                    self.addProperty(
                        pset=top,
                        prop=self.psetdefs[psetdef][i],
                        ptype=self.psetdefs[psetdef][i + 1],
                    )
        if idx != 0:
            # span top levels
            idx = self.propmodel.invisibleRootItem().index()
            for i in range(self.propmodel.rowCount()):
                if self.propmodel.item(i, 0).hasChildren():
                    self.form.treeProperties.setFirstColumnSpanned(i, idx, True)
            self.form.treeProperties.expandAll()
            self.form.comboPset.setCurrentIndex(0)

    def removeProperty(self):
        from PySide import QtGui

        sel = self.form.treeProperties.selectedIndexes()
        remove = []
        if sel:
            item = self.propmodel.itemFromIndex(sel[0])
            if item.toolTip() == "PropertySet":
                for i in range(item.rowCount()):
                    remove.append(item.child(i, 0).text() + ";;" + item.text())
                self.propmodel.takeRow(sel[0].row())
            else:
                pset = self.propmodel.itemFromIndex(sel[0].parent())
                itemtext = item.text()
                if isinstance(pset, QtGui.QStandardItem):
                    itemtext += ";;" + pset.text()
                remove.append(itemtext)
                pset.takeRow(sel[0].row())
            self.updateDicts(remove=remove)

    def onSelected(self, index):
        PARAMS.SetInt("IfcPropertiesSelectedState", getattr(index, "value", index))
        self.objectslist, searchterms = self.rebuildObjectsList()
        self.form.searchField.clear()
        self.form.searchField.addItems(searchterms)
        self.update()

    def onVisible(self, index):
        PARAMS.SetInt("IfcPropertiesVisibleState", getattr(index, "value", index))
        self.update()


if FreeCAD.GuiUp:
    from PySide import QtGui

    class propertiesDelegate(QtGui.QStyledItemDelegate):
        def __init__(self, parent=None, container=None, ptypes=[], plabels=[], *args):
            self.container = container
            QtGui.QStyledItemDelegate.__init__(self, parent, *args)
            self.ptypes = ptypes
            self.plabels = plabels

        def createEditor(self, parent, option, index):
            import FreeCADGui

            if index.column() == 0:  # property name
                editor = QtGui.QLineEdit(parent)
            elif index.column() == 1:  # property type
                editor = QtGui.QComboBox(parent)
            else:  # property value
                ptype = index.sibling(index.row(), 1).data()
                if "Integer" in ptype:
                    editor = QtGui.QSpinBox(parent)
                elif "Real" in ptype:
                    editor = QtGui.QDoubleSpinBox(parent)
                    editor.setDecimals(
                        FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Units").GetInt(
                            "Decimals", 2
                        )
                    )
                elif ("Boolean" in ptype) or ("Logical" in ptype):
                    editor = QtGui.QComboBox(parent)
                    editor.addItems(["True", "False"])
                elif "Measure" in ptype:
                    editor = FreeCADGui.UiLoader().createWidget("Gui::InputField")
                    editor.setParent(parent)
                else:
                    editor = QtGui.QLineEdit(parent)
                editor.setObjectName("editor_" + ptype)
            return editor

        def setEditorData(self, editor, index):
            if index.column() == 0:
                editor.setText(index.data())
            elif index.column() == 1:
                editor.addItems(self.plabels)
                if index.data() in self.plabels:
                    idx = self.plabels.index(index.data())
                    editor.setCurrentIndex(idx)
            else:
                if "Integer" in editor.objectName():
                    try:
                        editor.setValue(int(index.data()))
                    except (TypeError, ValueError, AttributeError):
                        editor.setValue(0)
                elif "Real" in editor.objectName():
                    try:
                        editor.setValue(float(index.data()))
                    except (TypeError, ValueError, AttributeError):
                        editor.setValue(0)
                elif ("Boolean" in editor.objectName()) or ("Logical" in editor.objectName()):
                    try:
                        editor.setCurrentIndex(["true", "false"].index(index.data().lower()))
                    except (ValueError, AttributeError):
                        editor.setCurrentIndex(1)
                elif "Measure" in editor.objectName():
                    try:
                        editor.setText(index.data())
                    except (ValueError, AttributeError):
                        editor.setValue(0)
                else:
                    editor.setText(index.data())

        def setModelData(self, editor, model, index):
            remove = []
            if index.column() == 0:
                oldtext = index.data()
                if oldtext != editor.text():
                    pset = index.parent().data()
                    remove.append(oldtext + ";;" + pset)
                basename = editor.text()
                # check for duplicate name
                names = []
                for i in range(model.rowCount()):
                    topitem = model.item(i, 0)
                    names.append(topitem.text())
                    if topitem.hasChildren():
                        for j in range(topitem.rowCount()):
                            childitem = topitem.child(j, 0)
                            names.append(childitem.text())
                suffix = 1
                newname = basename
                while newname in names:
                    newname = basename + str(suffix).zfill(3)
                    suffix += 1
                model.setData(index, newname)
            elif index.column() == 1:
                if editor.currentIndex() > -1:
                    idx = editor.currentIndex()
                    data = self.plabels[idx]
                    model.setData(index, data)
            else:
                if ("Integer" in editor.objectName()) or ("Real" in editor.objectName()):
                    model.setData(index, str(editor.value()))
                elif ("Boolean" in editor.objectName()) or ("Logical" in editor.objectName()):
                    model.setData(index, editor.currentText())
                elif "Measure" in editor.objectName():
                    model.setData(index, editor.property("text"))
                else:
                    model.setData(index, editor.text())
            self.container.updateDicts(remove)


FreeCADGui.addCommand("BIM_IfcProperties", BIM_IfcProperties())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2026 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "Abstract base class for the work with solvers and meshers"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"


from PySide.QtCore import QProcess
from abc import ABC, abstractmethod
import os
import tempfile

import FreeCAD


class ObjectTools(ABC):
    """Abstract base class for the work with solvers and meshers"""

    def __init__(self, obj):
        obj.Tool = self
        self.obj = obj
        self.process = QProcess()
        self.analysis = obj.getParentGroup()
        self.fem_param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem")
        self._create_working_directory(obj)

        self.process.finished.connect(self._process_finished)

    def _create_working_directory(self, obj):
        """
        Create working directory according to preferences
        """
        if not os.path.isdir(obj.WorkingDirectory):
            gen_param = self.fem_param.GetGroup("General")
            if gen_param.GetBool("UseTempDirectory"):
                self.obj.WorkingDirectory = tempfile.mkdtemp(prefix="fem_")
            elif gen_param.GetBool("UseBesideDirectory"):
                root, ext = os.path.splitext(obj.Document.FileName)
                if root:
                    self.obj.WorkingDirectory = os.path.join(root, obj.Label)
                    os.makedirs(self.obj.WorkingDirectory, exist_ok=True)
                else:
                    # file not saved, use temporary
                    self.obj.WorkingDirectory = tempfile.mkdtemp(prefix="fem_")
            elif gen_param.GetBool("UseCustomDirectory"):
                self.obj.WorkingDirectory = gen_param.GetString("CustomDirectoryPath")
                os.makedirs(self.obj.WorkingDirectory, exist_ok=True)

    @abstractmethod
    def prepare(self):
        pass

    @abstractmethod
    def compute(self):
        pass

    @abstractmethod
    def update_properties(self):
        pass

    def run(self, blocking=False):
        self.prepare()
        self.compute()
        if blocking:
            return self.process.waitForFinished(-1)
        return None

    def _process_finished(self, code, status):
        if status == QProcess.ExitStatus.NormalExit and code == 0:
            self.update_properties()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2022 - 2025 Larry Woestman <LarryWoestman2@gmail.com>   *
# *   Copyright (c) 2024 Ondsel <development@ondsel.com>                    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from typing import Any, Dict

from Path.Post.Processor import PostProcessor

import Path
import FreeCAD

translate = FreeCAD.Qt.translate

DEBUG = False
if DEBUG:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

#
# Define some types that are used throughout this file.
#
Values = Dict[str, Any]
Visible = Dict[str, bool]


class Refactored_Mach3_Mach4(PostProcessor):
    """The Refactored Mach3_Mach4 post processor class."""

    def __init__(
        self,
        job,
        tooltip=translate("CAM", "Refactored Mach3_Mach4 post processor"),
        tooltipargs=[""],
        units="Metric",
    ) -> None:
        super().__init__(
            job=job,
            tooltip=tooltip,
            tooltipargs=tooltipargs,
            units=units,
        )
        Path.Log.debug("Refactored Mach3_Mach4 post processor initialized.")

    def init_values(self, values: Values) -> None:
        """Initialize values that are used throughout the postprocessor."""
        #
        super().init_values(values)
        #
        # Set any values here that need to override the default values set
        # in the parent routine.
        #
        values["ENABLE_COOLANT"] = True
        #
        # Used in the argparser code as the "name" of the postprocessor program.
        # This would normally show up in the usage message in the TOOLTIP_ARGS.
        #
        values["MACHINE_NAME"] = "mach3_4"
        #
        # Enable special processing for operations with "Adaptive" in the name.
        #
        values["OUTPUT_ADAPTIVE"] = True
        #
        # Output the machine name for mach3_mach4 instead of the machine units alone.
        #
        values["OUTPUT_MACHINE_NAME"] = True
        #
        # The order of parameters.
        #
        # mach3_mach4 doesn't want K properties on XY plane; Arcs need work.
        #
        values["PARAMETER_ORDER"] = [
            "X",
            "Y",
            "Z",
            "A",
            "B",
            "C",
            "I",
            "J",
            "F",
            "S",
            "T",
            "Q",
            "R",
            "L",
            "H",
            "D",
            "P",
        ]
        #
        # Any commands in this value will be output as the last commands
        # in the G-code file.
        #
        values[
            "POSTAMBLE"
        ] = """M05
G17 G54 G90 G80 G40
M2"""
        values["POSTPROCESSOR_FILE_NAME"] = __name__
        #
        # Any commands in this value will be output after the header and
        # safety block at the beginning of the G-code file.
        #
        values["PREAMBLE"] = """G17 G54 G40 G49 G80 G90"""
        #
        # Output the machine name for mach3_mach4 instead of the machine units alone.
        #
        values["SHOW_MACHINE_UNITS"] = False

    def init_arguments_visible(self, arguments_visible: Visible) -> None:
        """Initialize which argument pairs are visible in TOOLTIP_ARGS."""
        super().init_arguments_visible(arguments_visible)
        #
        # Modify the visibility of any arguments from the defaults here.
        #
        arguments_visible["axis-modal"] = True

    @property
    def tooltip(self):
        tooltip: str = """
        This is a postprocessor file for the CAM workbench.
        It is used to take a pseudo-gcode fragment from a CAM object
        and output 'real' GCode suitable for a Mach3_4 3 axis mill.
        """
        return tooltip



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "NonGui Tutorial 01 - Eigenvalue of elastic beam",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": [],
        "solvers": ["ccxtools", "elmer"],
        "material": "solid",
        "equations": ["elasticity"],  # "frequency", but list not allowed here
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.elmer_nonguitutorial01_eigenvalue_of_elastic_beam import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?t=56590

"""
    )


def setup(doc=None, solvertype="elmer"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric object
    geom_obj = doc.addObject("Part::Box", "Box")
    geom_obj.Length = 1000
    geom_obj.Width = 200
    geom_obj.Height = 100
    doc.recompute()
    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    elif solvertype == "elmer":
        solver_obj = ObjectsFem.makeSolverElmer(doc, "SolverElmer")
        eq_obj = ObjectsFem.makeEquationElasticity(doc, solver_obj)
        eq_obj.EigenAnalysis = True
        eq_obj.CalculateStresses = True
        eq_obj.DisplaceMesh = False
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.AnalysisType = "frequency"
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
        solver_obj.EigenmodesCount = 5
        solver_obj.EigenmodeHighLimit = 1000000.0
        solver_obj.EigenmodeLowLimit = 0.01
    analysis.addObject(solver_obj)

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "MechanicalMaterial")
    mat = material_obj.Material
    mat["Name"] = "Steel-Generic"
    mat["YoungsModulus"] = "100 GPa"
    mat["PoissonRatio"] = "0.30"
    mat["Density"] = "2330 kg/m^3"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # constraint fixed
    con_fixed = ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    con_fixed.References = [(geom_obj, "Face1"), (geom_obj, "Face2")]
    analysis.addObject(con_fixed)

    # mesh
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False
    femmesh_obj.CharacteristicLengthMax = "40.80 mm"

    # generate the mesh
    success = generate_mesh.mesh_from_mesher(femmesh_obj, "gmsh")
    if not success:
        # try to create from existing mesh
        from .meshes.mesh_eigenvalue_of_elastic_beam_tetra10 import create_nodes, create_elements

        fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
        femmesh_obj.FemMesh = fem_mesh

    doc.recompute()
    return doc



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Class and methods to migrate old FEM Gui objects

see module end as well as forum topic
https://forum.freecad.org/viewtopic.php?&t=46218
"""

__title__ = "FEM GUI that migrates old scripted objects"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

from importlib.util import spec_from_loader


class FemMigrateGui:

    def find_spec(self, fullname, path, target=None):

        if fullname in {
            "femguiobjects",
            "femguiobjects._ViewProviderFemConstraintBodyHeatSource",
            "femguiobjects._ViewProviderFemConstraintElectrostaticPotential",
            "femguiobjects._ViewProviderFemConstraintFlowVelocity",
            "femguiobjects._ViewProviderFemConstraintInitialFlowVelocity",
            "femguiobjects._ViewProviderFemConstraintSelfWeight",
            "femguiobjects._ViewProviderFemConstraintTie",
            "femguiobjects._ViewProviderFemElementFluid1D",
            "femguiobjects._ViewProviderFemElementGeometry1D",
            "femguiobjects._ViewProviderFemElementGeometry2D",
            "femguiobjects._ViewProviderFemElementRotation1D",
            "femguiobjects._ViewProviderFemMaterial",
            "femguiobjects._ViewProviderFemMaterialMechanicalNonlinear",
            "femguiobjects._ViewProviderFemMaterialReinforced",
            "femguiobjects._ViewProviderFemMeshBoundaryLayer",
            "femguiobjects._ViewProviderFemMeshGmsh",
            "femguiobjects._ViewProviderFemMeshGroup",
            "femguiobjects._ViewProviderFemMeshRegion",
            "femguiobjects._ViewProviderFemMeshResult",
            "femguiobjects._ViewProviderFemResultMechanical",
            "femguiobjects._ViewProviderFemSolverCalculix",
            "PyGui",
            "PyGui._ViewProviderFemConstraintBodyHeatSource",
            "PyGui._ViewProviderFemConstraintElectrostaticPotential",
            "PyGui._ViewProviderFemConstraintFlowVelocity",
            "PyGui._ViewProviderFemConstraintSelfWeight",
            "PyGui._ViewProviderFemElementFluid1D",
            "PyGui._ViewProviderFemElementGeometry1D",
            "PyGui._ViewProviderFemElementGeometry2D",
            "PyGui._ViewProviderFemElementRotation1D",
            "PyGui._ViewProviderFemMaterial",
            "PyGui._ViewProviderFemMaterialMechanicalNonlinear",
            "PyGui._ViewProviderFemMeshBoundaryLayer",
            "PyGui._ViewProviderFemMeshGmsh",
            "PyGui._ViewProviderFemMeshGroup",
            "PyGui._ViewProviderFemMeshRegion",
            "PyGui._ViewProviderFemMeshResult",
            "PyGui._ViewProviderFemResultMechanical",
            "PyGui._ViewProviderFemSolverCalculix",
            "PyGui._ViewProviderFemSolverZ88",
            "PyGui._ViewProviderFemBeamSection",
            "PyGui._ViewProviderFemFluidSection",
            "PyGui._ViewProviderFemShellThickness",
            "_ViewProviderFemBeamSection",
            "_ViewProviderFemConstraintSelfWeight",
            "_ViewProviderFemMaterial",
            "_ViewProviderFemMaterialMechanicalNonlinear",
            "_ViewProviderFemMeshGmsh",
            "_ViewProviderFemMeshGroup",
            "_ViewProviderFemMeshRegion",
            "_ViewProviderFemResultMechanical",
            "_ViewProviderFemShellThickness",
            "_ViewProviderFemSolverCalculix",
            "_ViewProviderFemSolverZ88",
            "_ViewProviderFemMechanicalResult",
            "ViewProviderFemResult",
            "_ViewProviderMechanicalMaterial",
        }:
            return spec_from_loader(fullname, self)

        return None

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        return self.load_module(module)

    def load_module(self, module):

        if module.__name__ == "femguiobjects":
            module.__path__ = "femguiobjects"
        if module.__name__ == "femguiobjects._ViewProviderFemConstraintBodyHeatSource":
            import femviewprovider.view_constraint_bodyheatsource

            module.ViewProxy = (
                femviewprovider.view_constraint_bodyheatsource.VPConstraintBodyHeatSource
            )
        if module.__name__ == "femguiobjects._ViewProviderFemConstraintElectrostaticPotential":
            import femviewprovider.view_constraint_electrostaticpotential

            module.ViewProxy = (
                femviewprovider.view_constraint_electrostaticpotential.VPConstraintElectroStaticPotential
            )
        if module.__name__ == "femguiobjects._ViewProviderFemConstraintFlowVelocity":
            import femviewprovider.view_constraint_flowvelocity

            module.ViewProxy = femviewprovider.view_constraint_flowvelocity.VPConstraintFlowVelocity
        if module.__name__ == "femguiobjects._ViewProviderFemConstraintInitialFlowVelocity":
            import femviewprovider.view_constraint_initialflowvelocity

            module.ViewProxy = (
                femviewprovider.view_constraint_initialflowvelocity.VPConstraintInitialFlowVelocity
            )
        if module.__name__ == "femguiobjects._ViewProviderFemConstraintSelfWeight":
            import femviewprovider.view_constraint_selfweight

            module._ViewProviderFemConstraintSelfWeight = (
                femviewprovider.view_constraint_selfweight.VPConstraintSelfWeight
            )
        if module.__name__ == "femguiobjects._ViewProviderFemConstraintTie":
            import femviewprovider.view_constraint_tie

            module._ViewProviderFemConstraintTie = (
                femviewprovider.view_constraint_tie.VPConstraintTie
            )
        if module.__name__ == "femguiobjects._ViewProviderFemElementFluid1D":
            import femviewprovider.view_element_fluid1D

            module._ViewProviderFemElementFluid1D = (
                femviewprovider.view_element_fluid1D.VPElementFluid1D
            )
        if module.__name__ == "femguiobjects._ViewProviderFemElementGeometry1D":
            import femviewprovider.view_element_geometry1D

            module._ViewProviderFemElementGeometry1D = (
                femviewprovider.view_element_geometry1D.VPElementGeometry1D
            )
        if module.__name__ == "femguiobjects._ViewProviderFemElementGeometry2D":
            import femviewprovider.view_element_geometry2D

            module._ViewProviderFemElementGeometry2D = (
                femviewprovider.view_element_geometry2D.VPElementGeometry2D
            )
        if module.__name__ == "femguiobjects._ViewProviderFemElementRotation1D":
            import femviewprovider.view_element_rotation1D

            module._ViewProviderFemElementRotation1D = (
                femviewprovider.view_element_rotation1D.VPElementRotation1D
            )
        if module.__name__ == "femguiobjects._ViewProviderFemMaterial":
            import femviewprovider.view_material_common

            module._ViewProviderFemMaterial = femviewprovider.view_material_common.VPMaterialCommon
        if module.__name__ == "femguiobjects._ViewProviderFemMaterialMechanicalNonlinear":
            import femviewprovider.view_material_mechanicalnonlinear

            module._ViewProviderFemMaterialMechanicalNonlinear = (
                femviewprovider.view_material_mechanicalnonlinear.VPMaterialMechanicalNonlinear
            )
        if module.__name__ == "femguiobjects._ViewProviderFemMaterialReinforced":
            import femviewprovider.view_material_reinforced

            module._ViewProviderFemMaterialReinforced = (
                femviewprovider.view_material_reinforced.VPMaterialReinforced
            )
        if module.__name__ == "femguiobjects._ViewProviderFemMeshBoundaryLayer":
            import femviewprovider.view_mesh_boundarylayer

            module._ViewProviderFemMeshBoundaryLayer = (
                femviewprovider.view_mesh_boundarylayer.VPMeshBoundaryLayer
            )
        if module.__name__ == "femguiobjects._ViewProviderFemMeshGmsh":
            import femviewprovider.view_mesh_gmsh

            module._ViewProviderFemMeshGmsh = femviewprovider.view_mesh_gmsh.VPMeshGmsh
        if module.__name__ == "femguiobjects._ViewProviderFemMeshGroup":
            import femviewprovider.view_mesh_group

            module._ViewProviderFemMeshGroup = femviewprovider.view_mesh_group.VPMeshGroup
        if module.__name__ == "femguiobjects._ViewProviderFemMeshRegion":
            import femviewprovider.view_mesh_region

            module._ViewProviderFemMeshRegion = femviewprovider.view_mesh_region.VPMeshRegion
        if module.__name__ == "femguiobjects._ViewProviderFemMeshResult":
            import femviewprovider.view_mesh_result

            module._ViewProviderFemMeshResult = femviewprovider.view_mesh_result.VPFemMeshResult
        if module.__name__ == "femguiobjects._ViewProviderFemResultMechanical":
            import femviewprovider.view_result_mechanical

            module._ViewProviderFemResultMechanical = (
                femviewprovider.view_result_mechanical.VPResultMechanical
            )
        if module.__name__ == "femguiobjects._ViewProviderFemSolverCalculix":
            import femviewprovider.view_solver_ccxtools

            module._ViewProviderFemSolverCalculix = (
                femviewprovider.view_solver_ccxtools.VPSolverCcxTools
            )

        if module.__name__ == "PyGui":
            module.__path__ = "PyGui"
        if module.__name__ == "PyGui._ViewProviderFemConstraintBodyHeatSource":
            import femviewprovider.view_constraint_bodyheatsource

            module.ViewProxy = (
                femviewprovider.view_constraint_bodyheatsource.VPConstraintBodyHeatSource
            )
        if module.__name__ == "PyGui._ViewProviderFemConstraintElectrostaticPotential":
            import femviewprovider.view_constraint_electrostaticpotential

            module.ViewProxy = (
                femviewprovider.view_constraint_electrostaticpotential.VPConstraintElectroStaticPotential
            )
        if module.__name__ == "PyGui._ViewProviderFemConstraintFlowVelocity":
            import femviewprovider.view_constraint_flowvelocity

            module.ViewProxy = femviewprovider.view_constraint_flowvelocity.VPConstraintFlowVelocity
        if module.__name__ == "PyGui._ViewProviderFemConstraintInitialFlowVelocity":
            import femviewprovider.view_constraint_initialflowvelocity

            module.ViewProxy = (
                femviewprovider.view_constraint_initialflowvelocity.VPConstraintInitialFlowVelocity
            )
        if module.__name__ == "PyGui._ViewProviderFemConstraintSelfWeight":
            import femviewprovider.view_constraint_selfweight

            module._ViewProviderFemConstraintSelfWeight = (
                femviewprovider.view_constraint_selfweight.VPConstraintSelfWeight
            )
        if module.__name__ == "PyGui._ViewProviderFemElementFluid1D":
            import femviewprovider.view_element_fluid1D

            module._ViewProviderFemElementFluid1D = (
                femviewprovider.view_element_fluid1D.VPElementFluid1D
            )
        if module.__name__ == "PyGui._ViewProviderFemElementGeometry1D":
            import femviewprovider.view_element_geometry1D

            module._ViewProviderFemElementGeometry1D = (
                femviewprovider.view_element_geometry1D.VPElementGeometry1D
            )
        if module.__name__ == "PyGui._ViewProviderFemElementGeometry2D":
            import femviewprovider.view_element_geometry2D

            module._ViewProviderFemElementGeometry2D = (
                femviewprovider.view_element_geometry2D.VPElementGeometry2D
            )
        if module.__name__ == "PyGui._ViewProviderFemElementRotation1D":
            import femviewprovider.view_element_rotation1D

            module._ViewProviderFemElementRotation1D = (
                femviewprovider.view_element_rotation1D.VPElementRotation1D
            )
        if module.__name__ == "PyGui._ViewProviderFemMaterial":
            import femviewprovider.view_material_common

            module._ViewProviderFemMaterial = femviewprovider.view_material_common.VPMaterialCommon
        if module.__name__ == "PyGui._ViewProviderFemMaterialMechanicalNonlinear":
            import femviewprovider.view_material_mechanicalnonlinear

            module._ViewProviderFemMaterialMechanicalNonlinear = (
                femviewprovider.view_material_mechanicalnonlinear.VPMaterialMechanicalNonlinear
            )
        if module.__name__ == "PyGui._ViewProviderFemMeshBoundaryLayer":
            import femviewprovider.view_mesh_boundarylayer

            module._ViewProviderFemMeshBoundaryLayer = (
                femviewprovider.view_mesh_boundarylayer.VPMeshBoundaryLayer
            )
        if module.__name__ == "PyGui._ViewProviderFemMeshGmsh":
            import femviewprovider.view_mesh_gmsh

            module._ViewProviderFemMeshGmsh = femviewprovider.view_mesh_gmsh.VPMeshGmsh
        if module.__name__ == "PyGui._ViewProviderFemMeshGroup":
            import femviewprovider.view_mesh_group

            module._ViewProviderFemMeshGroup = femviewprovider.view_mesh_group.VPMeshGroup
        if module.__name__ == "PyGui._ViewProviderFemMeshRegion":
            import femviewprovider.view_mesh_region

            module._ViewProviderFemMeshRegion = femviewprovider.view_mesh_region.VPMeshRegion
        if module.__name__ == "PyGui._ViewProviderFemMeshResult":
            import femviewprovider.view_mesh_result

            module._ViewProviderFemMeshResult = femviewprovider.view_mesh_result.VPFemMeshResult
        if module.__name__ == "PyGui._ViewProviderFemResultMechanical":
            import femviewprovider.view_result_mechanical

            module._ViewProviderFemResultMechanical = (
                femviewprovider.view_result_mechanical.VPResultMechanical
            )
        if module.__name__ == "PyGui._ViewProviderFemSolverCalculix":
            import femviewprovider.view_solver_ccxtools

            module._ViewProviderFemSolverCalculix = (
                femviewprovider.view_solver_ccxtools.VPSolverCcxTools
            )
        if module.__name__ == "PyGui._ViewProviderFemSolverZ88":
            import femsolver.z88.solver

            module._ViewProviderFemSolverZ88 = femsolver.z88.solver.ViewProxy

        if module.__name__ == "PyGui._ViewProviderFemBeamSection":
            import femviewprovider.view_element_geometry1D

            module._ViewProviderFemBeamSection = (
                femviewprovider.view_element_geometry1D.VPElementGeometry1D
            )
        if module.__name__ == "PyGui._ViewProviderFemFluidSection":
            import femviewprovider.view_element_fluid1D

            module._ViewProviderFemFluidSection = (
                femviewprovider.view_element_fluid1D.VPElementFluid1D
            )
        if module.__name__ == "PyGui._ViewProviderFemShellThickness":
            import femviewprovider.view_element_geometry2D

            module._ViewProviderFemShellThickness = (
                femviewprovider.view_element_geometry2D.VPElementGeometry2D
            )

        if module.__name__ == "_ViewProviderFemBeamSection":
            import femviewprovider.view_element_geometry1D

            module._ViewProviderFemBeamSection = (
                femviewprovider.view_element_geometry1D.VPElementGeometry1D
            )
        if module.__name__ == "_ViewProviderFemConstraintSelfWeight":
            import femviewprovider.view_constraint_selfweight

            module._ViewProviderFemConstraintSelfWeight = (
                femviewprovider.view_constraint_selfweight.VPConstraintSelfWeight
            )
        if module.__name__ == "_ViewProviderFemMaterial":
            import femviewprovider.view_material_common

            module._ViewProviderFemMaterial = femviewprovider.view_material_common.VPMaterialCommon
        if module.__name__ == "_ViewProviderFemMaterialMechanicalNonlinear":
            import femviewprovider.view_material_mechanicalnonlinear

            module._ViewProviderFemMaterialMechanicalNonlinear = (
                femviewprovider.view_material_mechanicalnonlinear.VPMaterialMechanicalNonlinear
            )
        if module.__name__ == "_ViewProviderFemMeshGmsh":
            import femviewprovider.view_mesh_gmsh

            module._ViewProviderFemMeshGmsh = femviewprovider.view_mesh_gmsh.VPMeshGmsh
        if module.__name__ == "_ViewProviderFemMeshGroup":
            import femviewprovider.view_mesh_group

            module._ViewProviderFemMeshGroup = femviewprovider.view_mesh_group.VPMeshGroup
        if module.__name__ == "_ViewProviderFemMeshRegion":
            import femviewprovider.view_mesh_region

            module._ViewProviderFemMeshRegion = femviewprovider.view_mesh_region.VPMeshRegion
        if module.__name__ == "_ViewProviderFemResultMechanical":
            import femviewprovider.view_result_mechanical

            module._ViewProviderFemResultMechanical = (
                femviewprovider.view_result_mechanical.VPResultMechanical
            )
        if module.__name__ == "_ViewProviderFemShellThickness":
            import femviewprovider.view_element_geometry2D

            module._ViewProviderFemShellThickness = (
                femviewprovider.view_element_geometry2D.VPElementGeometry2D
            )
        if module.__name__ == "_ViewProviderFemSolverCalculix":
            import femviewprovider.view_solver_ccxtools

            module._ViewProviderFemSolverCalculix = (
                femviewprovider.view_solver_ccxtools.VPSolverCcxTools
            )
        if module.__name__ == "_ViewProviderFemSolverZ88":
            import femsolver.z88.solver

            module._ViewProviderFemSolverZ88 = femsolver.z88.solver.ViewProxy

        if module.__name__ == "_ViewProviderFemMechanicalResult":
            import femviewprovider.view_result_mechanical

            module._ViewProviderFemMechanicalResult = (
                femviewprovider.view_result_mechanical.VPResultMechanical
            )
        if module.__name__ == "ViewProviderFemResult":
            import femviewprovider.view_result_mechanical

            module.ViewProviderFemResult = femviewprovider.view_result_mechanical.VPResultMechanical
        if module.__name__ == "_ViewProviderMechanicalMaterial":
            import femviewprovider.view_material_common

            module._ViewProviderMechanicalMaterial = (
                femviewprovider.view_material_common.VPMaterialCommon
            )

        return None


"""
possible entries in the old files:
(the class name in the old file does not matter, we ever only had one class per module)

further renaming objects
module="femsolver.elmer.equations.fluxsolver"
see App migrate because object class and viewprovider class are in the same module

fourth big moving
renaming class and module names in femobjects
TODO add link to commit before the first commit
module="femguiobjects._ViewProviderFemConstraintBodyHeatSource"
module="femguiobjects._ViewProviderFemConstraintElectrostaticPotential"
module="femguiobjects._ViewProviderFemConstraintFlowVelocity"
module="femguiobjects._ViewProviderFemConstraintInitialFlowVelocity"
module="femguiobjects._ViewProviderFemConstraintSelfWeight"
module="femguiobjects._ViewProviderFemConstraintTie"
module="femguiobjects._ViewProviderFemElementFluid1D"
module="femguiobjects._ViewProviderFemElementGeometry1D"
module="femguiobjects._ViewProviderFemElementGeometry2D"
module="femguiobjects._ViewProviderFemElementRotation1D"
module="femguiobjects._ViewProviderFemMaterial"
module="femguiobjects._ViewProviderFemMaterialMechanicalNonlinear"
module="femguiobjects._ViewProviderFemMaterialReinforced"
module="femguiobjects._ViewProviderFemMeshBoundaryLayer"
module="femguiobjects._ViewProviderFemMeshGmsh"
module="femguiobjects._ViewProviderFemMeshGroup"
module="femguiobjects._ViewProviderFemMeshRegion"
module="femguiobjects._ViewProviderFemMeshResult"
module="femguiobjects._ViewProviderFemResultMechanical"
module="femguiobjects._ViewProviderFemSolverCalculix"

third big moving
from PyGui to femguiobjects, following the parent commit
https://github.com/FreeCAD/FreeCAD/tree/07ae0e56c4/src/Mod/Fem/PyGui
module="PyGui._ViewProviderFemConstraintBodyHeatSource"
module="PyGui._ViewProviderFemConstraintElectrostaticPotential"
module="PyGui._ViewProviderFemConstraintFlowVelocity"
module="PyGui._ViewProviderFemConstraintInitialFlowVelocity"
module="PyGui._ViewProviderFemConstraintSelfWeight"
module="PyGui._ViewProviderFemElementFluid1D"
module="PyGui._ViewProviderFemElementGeometry1D"
module="PyGui._ViewProviderFemElementGeometry2D"
module="PyGui._ViewProviderFemElementRotation1D"
module="PyGui._ViewProviderFemMaterial"
module="PyGui._ViewProviderFemMaterialMechanicalNonlinear"
module="PyGui._ViewProviderFemMeshBoundaryLayer"
module="PyGui._ViewProviderFemMeshGmsh"
module="PyGui._ViewProviderFemMeshGroup"
module="PyGui._ViewProviderFemMeshRegion"
module="PyGui._ViewProviderFemMeshResult"
module="PyGui._ViewProviderFemResultMechanical"
module="PyGui._ViewProviderFemSolverCalculix"
module="PyGui._ViewProviderFemSolverZ88"

renamed between the second and third big moveings
module="PyGui._ViewProviderFemBeamSection"
module="PyGui._ViewProviderFemFluidSection"
module="PyGui._ViewProviderFemShellThickness"

second big moveing
into PyObjects, following the parent commit
https://github.com/FreeCAD/FreeCAD/tree/7f884e8bff/src/Mod/Fem
module="_ViewProviderFemBeamSection"
module="_ViewProviderFemConstraintSelfWeight"
module="_ViewProviderFemMaterial"
module="_ViewProviderFemMaterialMechanicalNonlinear"
module="_ViewProviderFemMeshGmsh"
module="_ViewProviderFemMeshGroup"
module="_ViewProviderFemMeshRegion"
module="_ViewProviderFemResultMechanical"
module="_ViewProviderFemShellThickness"
module="_ViewProviderFemSolverCalculix"
module="_ViewProviderFemSolverZ88"

renamed between the first and second big moveings
module="_ViewProviderFemMechanicalResult"
module="ViewProviderFemResult"
module="_ViewProviderMechanicalMaterial"

first big moving
split modules from one module into make, obj class, vp class, command
new obj class module names had a _
following the parent commit of the first split commit
https://github.com/FreeCAD/FreeCAD/tree/c3328d6b4e/src/Mod/Fem
in this modules there where object class and viewprovider class together
# see migrate App
module="FemBeamSection"
module="FemShellThickness"
module="MechanicalAnalysis"
module="MechanicalMaterial"
"""



import FreeCAD
import os


def createPageWithSVGTemplate(doc=None):
    """Returns a page with an SVGTemplate added on the ActiveDocument"""
    path = os.path.dirname(os.path.abspath(__file__))
    templateFileSpec = path + "/TestTemplate.svg"

    if not doc:
        doc = FreeCAD.ActiveDocument

    page = doc.addObject("TechDraw::DrawPage", "Page")
    doc.addObject("TechDraw::DrawSVGTemplate", "Template")
    doc.Template.Template = templateFileSpec
    doc.Page.Template = doc.Template
    return page



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2022 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Elmer equation object _NonLinear"
__author__ = "Markus Hovorka, Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from . import linear


# the linear equation object defines some attributes for some various elmer equations
# these various elmer equations are based on the linear equation object
# thus in ObjectsFem module is no method to add a linear equation object


class Proxy(linear.Proxy):

    def __init__(self, obj):
        super().__init__(obj)

        obj.addProperty(
            "App::PropertyIntegerConstraint",
            "NonlinearIterations",
            "Nonlinear System",
            "Maximum number of iterations",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyIntegerConstraint",
            "NonlinearNewtonAfterIterations",
            "Nonlinear System",
            "",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyFloat",
            "NonlinearNewtonAfterTolerance",
            "Nonlinear System",
            "",
            locked=True,
        )
        obj.addProperty(
            "App::PropertyFloat", "NonlinearTolerance", "Nonlinear System", "", locked=True
        )
        obj.addProperty(
            "App::PropertyFloatConstraint",
            "RelaxationFactor",
            "Nonlinear System",
            (
                "Value below 1.0 might be necessary to achieve convergence\n"
                "Typical values are in the range [0.3, 1.0]"
            ),
            locked=True,
        )

        obj.NonlinearIterations = (20, 1, int(1e6), 10)
        obj.NonlinearNewtonAfterIterations = (3, 1, 500, 1)
        # for small numbers we must set an expression because we don't have a UI,
        # the user has to view and edit the tolerance via the property editor and
        # this does not yet allow one to view and edit small numbers in scientific notation
        # forum thread: https://forum.freecad.org/viewtopic.php?p=613897#p613897
        obj.setExpression("NonlinearTolerance", "1e-7")
        obj.setExpression("NonlinearNewtonAfterTolerance", "1e-3")
        obj.RelaxationFactor = (1.0, 0.0001, 2.0, 0.1)


class ViewProxy(linear.ViewProxy):
    pass


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import PathScripts
from PySide.QtCore import QT_TRANSLATE_NOOP

__doc__ = """CAM SimpleCopy command"""

translate = FreeCAD.Qt.translate


class ViewProvider:

    def __init__(self, vobj):
        self.Object = vobj.Object
        vobj.Proxy = self

    def attach(self, vobj):
        self.Object = vobj.Object
        return

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onChanged(self, vobj, prop):
        return

    def getIcon(self):
        if self.Object.Active:
            return ":/icons/CAM_SimpleCopy.svg"
        else:
            return ":/icons/CAM_OpActive.svg"


class CommandPathSimpleCopy:
    def GetResources(self):
        return {
            "Pixmap": "CAM_SimpleCopy",
            "MenuText": QT_TRANSLATE_NOOP("CAM_SimpleCopy", "Simple Copy"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_SimpleCopy", "Creates a non-parametric copy of another toolpath"
            ),
        }

    def IsActive(self):
        if bool(FreeCADGui.Selection.getSelection()) is False:
            return False
        try:
            obj = FreeCADGui.Selection.getSelectionEx()[0].Object
            return isinstance(obj.Proxy, Path.Op.Base.ObjectOp)
        except Exception:
            return False

    def Activated(self):
        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            FreeCAD.Console.PrintError(
                translate("CAM_SimpleCopy", "Select exactly one toolpath object") + "\n"
            )
            return
        if not (selection[0].isDerivedFrom("Path::Feature")):
            FreeCAD.Console.PrintError(
                translate("CAM_SimpleCopy", "Select exactly one toolpath object") + "\n"
            )
            return

        FreeCAD.ActiveDocument.openTransaction("Simple Copy")
        FreeCADGui.doCommand("srcobj = FreeCADGui.Selection.getSelectionEx()[0].Object\n")

        FreeCADGui.addModule("PathScripts.PathUtils")
        FreeCADGui.doCommand("srcpath = PathScripts.PathUtils.getPathWithPlacement(srcobj)\n")
        FreeCADGui.addModule("Path.Op.Custom")
        FreeCADGui.doCommand('obj = Path.Op.Custom.Create("' + selection[0].Name + '_SimpleCopy")')
        FreeCADGui.doCommand("obj.Gcode = [c.toGCode() for c in srcpath.Commands]")
        FreeCADGui.doCommand(
            "obj.ViewObject.Proxy = Path.Op.Gui.SimpleCopy.ViewProvider(obj.ViewObject)"
        )
        FreeCADGui.doCommand("PathScripts.PathUtils.addToJob(obj)")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_SimpleCopy", CommandPathSimpleCopy())



# ***************************************************************************
# *   Copyright (c) 2019 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Import FEM unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import unittest

import FreeCAD

from .support_utils import fcc_print


class TestFemImport(unittest.TestCase):
    fcc_print("import TestFemImport")

    # ********************************************************************************************
    # no is document needed to test import Fem and import FemGui
    # thus neither setUp nor tearDown methods are needed

    def test_00print(self):
        fcc_print(
            "\n{0}\n{1} run FEM TestFemImport tests {2}\n{0}".format(100 * "*", 10 * "*", 61 * "*")
        )

    # ********************************************************************************************
    def test_import_fem(self):

        mod = "Fem"
        fcc_print(f"\n  Try importing {mod} ...")
        try:
            im = __import__(f"{mod}")
        except ImportError:
            im = False
        if not im:
            # to get an error message what was going wrong
            __import__(f"{mod}")
        self.assertTrue(im, f"Problem importing {mod}")

        if FreeCAD.GuiUp:
            mod = "FemGui"
            fcc_print(f"  Try importing {mod} ...")
            try:
                im = __import__(f"{mod}")
            except ImportError:
                im = False
            if not im:
                # to get an error message what was going wrong
                __import__(f"{mod}")
            self.assertTrue(im, f"Problem importing {mod}")


# ************************************************************************************************
# ************************************************************************************************
# to be sure this is run on very first of FEM test this is here and not in objects
class TestObjectExistance(unittest.TestCase):
    fcc_print("import TestObjectExistance")

    # ********************************************************************************************
    def setUp(self):
        # setUp is executed before every test

        # new document
        self.document = FreeCAD.newDocument(self.__class__.__name__)

    # ********************************************************************************************
    def tearDown(self):
        # tearDown is executed after every test
        FreeCAD.closeDocument(self.document.Name)

    # ********************************************************************************************
    def test_00print(self):
        # since method name starts with 00 this will be run first
        # this test just prints a line with stars
        fcc_print(
            "\n{0}\n{1} run FEM TestObjectExistance tests {2}\n{0}".format(
                100 * "*", 10 * "*", 55 * "*"
            )
        )

    # ********************************************************************************************
    def test_objects_existance(self):

        expected_obj_types = [
            "Fem::Constraint",
            "Fem::ConstraintBearing",
            "Fem::ConstraintContact",
            "Fem::ConstraintDisplacement",
            "Fem::ConstraintFixed",
            "Fem::ConstraintFluidBoundary",
            "Fem::ConstraintForce",
            "Fem::ConstraintGear",
            "Fem::ConstraintHeatflux",
            "Fem::ConstraintInitialTemperature",
            "Fem::ConstraintPlaneRotation",
            "Fem::ConstraintPressure",
            "Fem::ConstraintPulley",
            "Fem::ConstraintPython",
            "Fem::ConstraintTemperature",
            "Fem::ConstraintTransform",
            "Fem::DocumentObject",
            "Fem::FeaturePython",
            "Fem::FemAnalysis",
            "Fem::FemAnalysisPython",
            "Fem::FemMeshObject",
            "Fem::FemMeshObjectPython",
            "Fem::FemMeshShapeBaseObjectPython",
            "Fem::FemMeshShapeNetgenObject",
            "Fem::FemMeshShapeObject",
            "Fem::FemResultObject",
            "Fem::FemResultObjectPython",
            "Fem::FemSetElementsObject",
            "Fem::FemSetFacesObject",
            "Fem::FemSetGeometryObject",
            "Fem::FemSetNodesObject",
            "Fem::FemSetObject",
            "Fem::FemSolverObject",
            "Fem::FemSolverObjectPython",
        ]

        expected_vtk_obj_types = [
            "Fem::FemPostClipFilter",
            "Fem::FemPostContoursFilter",
            "Fem::FemPostCutFilter",
            "Fem::FemPostDataAlongLineFilter",
            "Fem::FemPostDataAtPointFilter",
            "Fem::FemPostFilter",
            "Fem::FemPostFunction",
            "Fem::FemPostFunctionProvider",
            "Fem::FemPostObject",
            "Fem::FemPostPipeline",
            "Fem::FemPostPlaneFunction",
            "Fem::FemPostScalarClipFilter",
            "Fem::FemPostSphereFunction",
            "Fem::FemPostWarpVectorFilter",
        ]

        # if FEM VTK post processing is enabled, we need to add VTK post objects
        if "BUILD_FEM_VTK" in FreeCAD.__cmake__:
            expected_obj_types += expected_vtk_obj_types

        expected_len = len(expected_obj_types)
        expected_obj_types = sorted(expected_obj_types)

        doc = self.document

        # get the supportedTypes for FEM module

        # Fem needs do be imported to get the FEM document types
        # with the following instead of import Fem
        # flake8 and lgtm do not complain "Fem imported but unused"
        __import__("Fem")

        obj_types = []
        for obj_type in sorted(doc.supportedTypes()):
            if obj_type.startswith("Fem"):
                obj_types.append(obj_type)

        obj_types = sorted(obj_types)

        # test
        self.assertEqual(expected_len, len(obj_types))

        self.assertEqual(expected_obj_types, obj_types)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 Werner Mayer <wmayer[at]users.sourceforge.net>     *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Surface unit tests"
__author__ = "Werner Mayer"
__url__ = "https://www.freecad.org"

import sys
import unittest
from os.path import join

import FreeCAD
from FreeCAD import Base
import Surface

vec = Base.Vector


class TestBlendCurve(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def assertAlmostCoincide(self, pt1, pt2, places=None, msg=None, delta=None):
        self.assertAlmostEqual(pt1.x, pt2.x, places, msg, delta)
        self.assertAlmostEqual(pt1.y, pt2.y, places, msg, delta)
        self.assertAlmostEqual(pt1.z, pt2.z, places, msg, delta)

    def test_blend_curve(self):
        # Create C0 BlendPoint at origin
        b1 = Surface.BlendPoint()
        # Create G1 BlendPoint
        b2 = Surface.BlendPoint([vec(10, 3, 6), vec(2, 5, 6)])
        # BlendCurve between the two BlendPoints
        bc = Surface.BlendCurve(b1, b2)
        # Compute the interpolating BezierCurve
        curve1 = bc.compute()

        # Create G2 BlendPoint at the end of previous BlendCurve
        b1 = Surface.BlendPoint(curve1.toShape(), 1, 2)
        # Create G1 BlendPoint
        b2 = Surface.BlendPoint([vec(5, 6, 2), vec(2, 5, 6)])

        bc = Surface.BlendCurve(b1, b2)
        # Compute the interpolating BezierCurve
        curve2 = bc.compute()

        d1 = curve1.getD2(1)
        d2 = curve2.getD2(0)

        self.assertEqual(len(d1), len(d2))
        self.assertAlmostCoincide(d1[0], d2[0])
        self.assertAlmostCoincide(d1[1], d2[1])
        self.assertAlmostCoincide(d1[2], d2[2])



# SPDX-License-Identifier: LGPL-2.1-or-later

# *****************************************************************************
# *                                                                           *
# *   Copyright (c) 2014 Jonathan Wiedemann <wood.galaxy@gmail.com> (cutplan) *
# *   Copyright (c) 2019 Jerome Laverroux <jerome.laverroux@free.fr> (cutline)*
# *   Copyright (c) 2023 FreeCAD Project Association                          *
# *                                                                           *
# *   This file is part of FreeCAD.                                           *
# *                                                                           *
# *   FreeCAD is free software: you can redistribute it and/or modify it      *
# *   under the terms of the GNU Lesser General Public License as             *
# *   published by the Free Software Foundation, either version 2.1 of the    *
# *   License, or (at your option) any later version.                         *
# *                                                                           *
# *   FreeCAD is distributed in the hope that it will be useful, but          *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of              *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU        *
# *   Lesser General Public License for more details.                         *
# *                                                                           *
# *   You should have received a copy of the GNU Lesser General Public        *
# *   License along with FreeCAD. If not, see                                 *
# *   <https://www.gnu.org/licenses/>.                                        *
# *                                                                           *
# *****************************************************************************

"""The Arch CutPlane command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_CutPlane:
    "the Arch CutPlane command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_CutPlane",
            "MenuText": QT_TRANSLATE_NOOP("Arch_CutPlane", "Cut With Plane"),
            "ToolTip": QT_TRANSLATE_NOOP("Arch_CutPlane", "Cut an object with a plane"),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and len(FreeCADGui.Selection.getSelection()) > 1

    def Activated(self):
        import ArchCutPlane

        baseObj, baseShp, cutterShp = ArchCutPlane._getShapes(
            FreeCADGui.Selection.getSelectionEx("", 0)
        )
        if baseObj is None:
            FreeCAD.Console.PrintError(
                translate(
                    "Arch",
                    "Select two objects, an object to be cut and an object defining a cutting plane, in that order",
                )
                + "\n"
            )
            return
        if baseShp is None:
            FreeCAD.Console.PrintError(
                translate("Arch", "The first object does not have a shape") + "\n"
            )
            return
        if cutterShp is None:
            FreeCAD.Console.PrintError(
                translate("Arch", "The second object does not define a plane") + "\n"
            )
            return
        panel = CutPlaneTaskPanel()
        FreeCADGui.Control.showDialog(panel)


class CutPlaneTaskPanel:
    def __init__(self):
        import ArchCutPlane
        from PySide import QtCore, QtGui

        _, self.base, self.cutter = ArchCutPlane._getShapes(
            FreeCADGui.Selection.getSelectionEx("", 0)
        )

        self.previewObj = FreeCAD.ActiveDocument.addObject("Part::Feature", "PreviewCutVolume")
        self.previewObj.ViewObject.ShapeColor = (1.00, 0.00, 0.00)
        self.previewObj.ViewObject.Transparency = 75

        self.form = QtGui.QWidget()
        self.form.setObjectName("TaskPanel")
        self.grid = QtGui.QGridLayout(self.form)
        self.grid.setObjectName("grid")
        self.title = QtGui.QLabel(self.form)
        self.grid.addWidget(self.title, 1, 0)
        self.infoText = QtGui.QLabel(self.form)
        self.grid.addWidget(self.infoText, 2, 0)
        self.combobox = QtGui.QComboBox()
        self.combobox.setCurrentIndex(0)
        self.grid.addWidget(self.combobox, 2, 1)
        QtCore.QObject.connect(
            self.combobox, QtCore.SIGNAL("currentIndexChanged(int)"), self.previewCutVolume
        )
        self.retranslateUi(self.form)
        self.previewCutVolume(self.combobox.currentIndex())

    def isAllowedAlterSelection(self):
        return False

    def accept(self):
        FreeCAD.ActiveDocument.removeObject(self.previewObj.Name)
        side = self.combobox.currentIndex()
        FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Cutting"))
        FreeCADGui.addModule("ArchCutPlane")
        FreeCADGui.doCommand("sels = FreeCADGui.Selection.getSelectionEx('', 0)")
        FreeCADGui.doCommand("ArchCutPlane.cutComponentwithPlane(sels, side=" + str(side) + ")")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()
        return True

    def reject(self):
        FreeCAD.ActiveDocument.removeObject(self.previewObj.Name)
        FreeCAD.Console.PrintMessage("Cancel Cut Plane\n")
        return True

    def getStandardButtons(self):
        from PySide import QtGui

        return QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel

    def previewCutVolume(self, i):
        import Arch

        cutVolume = Arch.getCutVolume(self.cutter, self.base)
        if i == 1:
            cutVolume = cutVolume[1]
        else:
            cutVolume = cutVolume[2]
        if cutVolume:
            self.previewObj.Shape = cutVolume

    def retranslateUi(self, TaskPanel):
        TaskPanel.setWindowTitle(translate("Arch", "Cut Plane"))
        self.title.setText(translate("Arch", "Cut Plane Options"))
        self.infoText.setText(translate("Arch", "Which side to cut"))
        self.combobox.addItems([translate("Arch", "Behind"), translate("Arch", "Front")])


FreeCADGui.addCommand("Arch_CutPlane", Arch_CutPlane())



# SPDX-License-Identifier: LGPL-2.1-or-later

# /***************************************************************************
#  *   Copyright (c) 2023 David Friedli <david[at]friedli-be.ch>             *
#  *                                                                         *
#  *   This file is part of FreeCAD.                                         *
#  *                                                                         *
#  *   FreeCAD is free software: you can redistribute it and/or modify it    *
#  *   under the terms of the GNU Lesser General Public License as           *
#  *   published by the Free Software Foundation, either version 2.1 of the  *
#  *   License, or (at your option) any later version.                       *
#  *                                                                         *
#  *   FreeCAD is distributed in the hope that it will be useful, but        *
#  *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
#  *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
#  *   Lesser General Public License for more details.                       *
#  *                                                                         *
#  *   You should have received a copy of the GNU Lesser General Public      *
#  *   License along with FreeCAD. If not, see                               *
#  *   <https://www.gnu.org/licenses/>.                                      *
#  *                                                                         *
#  **************************************************************************/


# Import the measure module here in order to ensure that default measurement types are loaded during startup.
# Note that they won't auto-load in gui-less mode. Ideally they would be loaded from "Init.py", similar
# to how the import/export types are registered. This would require to register measurement types from
# python which could complicate things further.

import Measure
import MeasureGui
from MeasureCOM import makeMeasureCOM, MeasureCOM


# Expose create functions
Measure.makeMeasureCOM = makeMeasureCOM


# Register python measure types
import FreeCAD
from PySide.QtCore import QT_TRANSLATE_NOOP


FreeCAD.MeasureManager.addMeasureType(
    "CENTEROFMASS",
    QT_TRANSLATE_NOOP("TaskMeasure", "Center of mass"),
    MeasureCOM,
)



# ***************************************************************************
# *   Copyright (c) 2015 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver calculix ccx tools task panel for the document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package task_solver_ccxtools
#  \ingroup FEM
#  \brief task panel for solver ccx tools object

import os
import time
from PySide import QtCore
from PySide import QtGui
from PySide.QtCore import Qt
from PySide.QtGui import QApplication

import FreeCAD
import FreeCADGui

import FemGui
from femtools.femutils import getOutputWinColor


def unicode(text, *args):
    return str(text)


class _TaskPanel:
    """
    The TaskPanel for CalculiX ccx tools solver object
    """

    PREFS_PATH = "User parameter:BaseApp/Preferences/Mod/Fem/Ccx"

    def __init__(self, solver_object):
        self.form = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/SolverCcxTools.ui"
        )

        from femtools.ccxtools import CcxTools as ccx

        # we do not need to pass the analysis, it will be found on fea init
        # TODO: if there is not analysis object in document init of fea
        # will fail with an exception and task panel will not open
        # handle more smart by a pop up error message and still open
        # task panel, may be deactivate write and run button.
        self.fea = ccx(solver_object)
        self.fea.setup_working_dir()
        try:
            self.fea.setup_ccx()
        except FileNotFoundError as e:
            FreeCAD.Console.PrintWarning(e.args[0])

        self.Calculix = QtCore.QProcess()
        self.Timer = QtCore.QTimer()
        self.Timer.start(300)

        self.fem_console_message = ""

        self.CCX_pipeline = None
        self.CCX_mesh_visibility = False

        # store visibility of possibly existing mesh object
        CCX_mesh = self.fea.analysis.Document.getObject("CCX_Results_Mesh")
        if CCX_mesh is not None:
            self.CCX_mesh_visibility = CCX_mesh.ViewObject.Visibility

        # Connect Signals and Slots
        QtCore.QObject.connect(
            self.form.tb_choose_working_dir, QtCore.SIGNAL("clicked()"), self.choose_working_dir
        )
        QtCore.QObject.connect(
            self.form.pb_write_inp, QtCore.SIGNAL("clicked()"), self.write_input_file_handler
        )
        QtCore.QObject.connect(
            self.form.pb_edit_inp, QtCore.SIGNAL("clicked()"), self.editCalculixInputFile
        )
        # connect stopCalculix before runCalculix
        # see https://github.com/FreeCAD/FreeCAD/issues/12448
        QtCore.QObject.connect(self.form.pb_run_ccx, QtCore.SIGNAL("clicked()"), self.stopCalculix)
        QtCore.QObject.connect(self.form.pb_run_ccx, QtCore.SIGNAL("clicked()"), self.runCalculix)
        QtCore.QObject.connect(
            self.form.rb_static_analysis, QtCore.SIGNAL("clicked()"), self.select_static_analysis
        )
        QtCore.QObject.connect(
            self.form.rb_frequency_analysis,
            QtCore.SIGNAL("clicked()"),
            self.select_frequency_analysis,
        )
        QtCore.QObject.connect(
            self.form.rb_thermomech_analysis,
            QtCore.SIGNAL("clicked()"),
            self.select_thermomech_analysis,
        )
        QtCore.QObject.connect(
            self.form.rb_check_mesh, QtCore.SIGNAL("clicked()"), self.select_check_mesh
        )
        QtCore.QObject.connect(
            self.form.rb_buckling_analysis,
            QtCore.SIGNAL("clicked()"),
            self.select_buckling_analysis,
        )
        QtCore.QObject.connect(self.Calculix, QtCore.SIGNAL("started()"), self.calculixStarted)
        QtCore.QObject.connect(
            self.Calculix,
            QtCore.SIGNAL("stateChanged(QProcess::ProcessState)"),
            self.calculixStateChanged,
        )
        QtCore.QObject.connect(
            self.Calculix, QtCore.SIGNAL("error(QProcess::ProcessError)"), self.calculixError
        )
        QtCore.QObject.connect(
            self.Calculix,
            QtCore.SIGNAL("finished(int, QProcess::ExitStatus)"),
            self.calculixFinished,
        )
        QtCore.QObject.connect(self.Timer, QtCore.SIGNAL("timeout()"), self.UpdateText)

        self.update()

    def getStandardButtons(self):
        # only show a close button
        # def accept() in no longer needed, since there is no OK button
        return QtGui.QDialogButtonBox.Close

    def reject(self):
        FreeCADGui.ActiveDocument.resetEdit()

    def update(self):
        "fills the widgets"
        self.form.le_working_dir.setText(self.fea.working_dir)
        if self.fea.solver.AnalysisType == "static":
            self.form.rb_static_analysis.setChecked(True)
        elif self.fea.solver.AnalysisType == "frequency":
            self.form.rb_frequency_analysis.setChecked(True)
        elif self.fea.solver.AnalysisType == "thermomech":
            self.form.rb_thermomech_analysis.setChecked(True)
        elif self.fea.solver.AnalysisType == "check":
            self.form.rb_check_mesh.setChecked(True)
        elif self.fea.solver.AnalysisType == "buckling":
            self.form.rb_buckling_analysis.setChecked(True)
        return

    def femConsoleMessage(self, message="", outputwin_color_type=None):
        self.fem_console_message = self.fem_console_message + (
            '<font color="{}"><b>{:4.1f}:</b></font> '.format(
                getOutputWinColor("Logging"), time.time() - self.Start
            )
        )
        if outputwin_color_type:
            if outputwin_color_type == "#00AA00":  # Success is not part of output window parameters
                self.fem_console_message += '<font color="{}">{}</font><br>'.format(
                    outputwin_color_type, message
                )
            else:
                self.fem_console_message += '<font color="{}">{}</font><br>'.format(
                    getOutputWinColor(outputwin_color_type), message
                )
        else:
            self.fem_console_message += message + "<br>"
        self.form.textEdit_Output.setText(self.fem_console_message)
        self.form.textEdit_Output.moveCursor(QtGui.QTextCursor.End)

    def printCalculiXstdout(self):

        out = self.Calculix.readAllStandardOutput()
        # print(type(out))
        # <class 'PySide2.QtCore.QByteArray'>

        if out.isEmpty():
            self.femConsoleMessage("CalculiX stdout is empty", "Error")
            return False

        # https://forum.freecad.org/viewtopic.php?f=18&t=39195
        # convert QByteArray to a binary string an decode it to "utf-8"
        out = out.data().decode()  # "utf-8" can be omitted
        out = os.linesep.join([s for s in out.splitlines() if s])
        out = out.replace("\n", "<br>")
        # print(out)
        self.femConsoleMessage(out)

        if "*ERROR in e_c3d: nonpositive jacobian" in out:
            error_message = (
                "\n\nCalculiX returned an error due to "
                "nonpositive jacobian determinant in at least one element\n"
                "Use the run button on selected solver to get a better error output.\n"
            )
            FreeCAD.Console.PrintError(error_message)

        if "*ERROR" in out:
            return False
        else:
            return True

    def UpdateText(self):
        if self.Calculix.state() == QtCore.QProcess.ProcessState.Running:
            self.form.l_time.setText(f"Time: {time.time() - self.Start:4.1f}: ")

    def calculixError(self, error=""):
        print(f"Error() {error}")
        self.femConsoleMessage(f"CalculiX execute error: {error}", "Error")

    def calculixNoError(self):
        print("CalculiX done without error!")
        self.femConsoleMessage(
            "CalculiX done without error!", "#00AA00"
        )  # Green leaving hard coded

    def calculixStarted(self):
        # print("calculixStarted()")
        FreeCAD.Console.PrintLog(f"calculix state: {self.Calculix.state()}\n")
        self.form.pb_run_ccx.setText("Stop CalculiX")

    def calculixStateChanged(self, newState):
        if newState == QtCore.QProcess.ProcessState.Starting:
            self.femConsoleMessage("Starting CalculiX...")
        elif newState == QtCore.QProcess.ProcessState.Running:
            self.femConsoleMessage("CalculiX is running...")
        elif newState == QtCore.QProcess.ProcessState.NotRunning:
            self.femConsoleMessage("CalculiX stopped.")
        else:
            self.femConsoleMessage("Problems.", "Error")

    def calculixFinished(self, exitCode, exitStatus):
        # print("calculixFinished(), exit code: {}".format(exitCode))
        FreeCAD.Console.PrintLog(f"calculix state: {self.Calculix.state()}\n")

        # Restore previous cwd
        QtCore.QDir.setCurrent(self.cwd)

        self.Timer.stop()

        self.form.pb_run_ccx.setText("Re-run CalculiX")

        if exitStatus != QtCore.QProcess.ExitStatus.NormalExit:
            return

        if self.printCalculiXstdout():
            self.calculixNoError()
        else:
            self.calculixError()

        self.femConsoleMessage("Loading result sets...")
        self.form.l_time.setText(f"Time: {time.time() - self.Start:4.1f}: ")
        self.fea.reset_mesh_purge_results_checked()
        self.fea.inp_file_name = self.fea.inp_file_name

        # check if ccx is greater than 2.10, if not do not read results
        # https://forum.freecad.org/viewtopic.php?f=18&t=23548#p183829 Point 3
        # https://forum.freecad.org/viewtopic.php?f=18&t=23548&start=20#p183909
        # https://forum.freecad.org/viewtopic.php?f=18&t=23548&start=30#p185027
        # https://github.com/FreeCAD/FreeCAD/commit/3dd1c9f
        majorVersion, minorVersion = self.fea.get_ccx_version()
        if majorVersion == 2 and minorVersion <= 10:
            message = (
                "The used CalculiX version {}.{} creates broken output files. "
                "The result file will not be read by FreeCAD FEM. "
                "You still can try to read it stand alone with FreeCAD, but it is "
                "strongly recommended to upgrade CalculiX to a newer version.\n".format(
                    majorVersion, minorVersion
                )
            )
            QtGui.QMessageBox.warning(None, "Upgrade CalculiX", message)
            raise

        QApplication.setOverrideCursor(Qt.WaitCursor)
        try:
            self.fea.load_results()
        except Exception as e:
            FreeCAD.Console.PrintError("loading results failed\n")
            FreeCAD.Console.PrintError(e)

        QApplication.restoreOverrideCursor()
        self.form.l_time.setText(f"Time: {time.time() - self.Start:4.1f}: ")

        # restore mesh object visibility
        CCX_mesh = self.fea.analysis.Document.getObject("ResultMesh")
        if CCX_mesh is not None:
            CCX_mesh.ViewObject.Visibility = self.CCX_mesh_visibility

    def choose_working_dir(self):
        wd = QtGui.QFileDialog.getExistingDirectory(
            None, "Choose CalculiX working directory", self.fea.working_dir
        )
        if os.path.isdir(wd):
            self.fea.setup_working_dir(wd)
        self.form.le_working_dir.setText(self.fea.working_dir)

    def write_input_file_handler(self):
        self.Start = time.time()
        self.form.l_time.setText(f"Time: {time.time() - self.Start:4.1f}: ")
        QApplication.restoreOverrideCursor()
        if self.check_prerequisites_helper():
            QApplication.setOverrideCursor(Qt.WaitCursor)
            self.fea.write_inp_file()
            if self.fea.inp_file_name != "":
                self.femConsoleMessage("Write completed.")
                self.form.pb_edit_inp.setEnabled(True)
                self.form.pb_run_ccx.setEnabled(True)
            else:
                self.femConsoleMessage("Write .inp file failed!", "Error")
            QApplication.restoreOverrideCursor()
        self.form.l_time.setText(f"Time: {time.time() - self.Start:4.1f}: ")

    def check_prerequisites_helper(self):
        self.Start = time.time()
        self.femConsoleMessage("Check dependencies...")
        self.form.l_time.setText(f"Time: {time.time() - self.Start:4.1f}: ")

        self.fea.update_objects()
        message = self.fea.check_prerequisites()
        if message != "":
            QtGui.QMessageBox.critical(None, "Missing prerequisite(s)", message)
            return False
        return True

    def start_ext_editor(self, ext_editor_path, filename):
        if not hasattr(self, "ext_editor_process"):
            self.ext_editor_process = QtCore.QProcess()
        if self.ext_editor_process.state() != QtCore.QProcess.Running:
            self.ext_editor_process.start(ext_editor_path, [filename])

    def editCalculixInputFile(self):
        print(f"editCalculixInputFile {self.fea.inp_file_name}")
        ccx_prefs = FreeCAD.ParamGet(self.PREFS_PATH)
        if ccx_prefs.GetBool("UseInternalEditor", True):
            FemGui.open(self.fea.inp_file_name)
        else:
            ext_editor_path = ccx_prefs.GetString("ExternalEditorPath", "")
            if ext_editor_path:
                self.start_ext_editor(ext_editor_path, self.fea.inp_file_name)
            else:
                print(
                    "External editor is not defined in FEM preferences. "
                    "Falling back to internal editor"
                )
                FemGui.open(self.fea.inp_file_name)

    def runCalculix(self):
        if self.Calculix.state() == QtCore.QProcess.ProcessState.NotRunning:
            if self.fea.ccx_binary_present is False:
                self.femConsoleMessage(
                    "CalculiX can not be started. Missing or incorrect CalculiX binary: {}".format(
                        self.fea.ccx_binary
                    )
                )
                # TODO deactivate the run button
                return
            # print("runCalculix")
            self.Start = time.time()

            self.femConsoleMessage(f"CalculiX binary: {self.fea.ccx_binary}")
            self.femConsoleMessage(f"CalculiX input file: {self.fea.inp_file_name}")
            self.femConsoleMessage("Run CalculiX...")

            FreeCAD.Console.PrintMessage(
                f"run CalculiX at: {self.fea.ccx_binary} with: {self.fea.inp_file_name}\n"
            )
            # change cwd because ccx may crash if directory has no write permission
            # there is also a limit of the length of file names so jump to the document directory

            # Set up for multi-threading. Note: same functionality as ccx_tools.py/start_ccx()
            ccx_prefs = FreeCAD.ParamGet(self.PREFS_PATH)
            env = QtCore.QProcessEnvironment.systemEnvironment()
            num_cpu_pref = ccx_prefs.GetInt("AnalysisNumCPUs", QtCore.QThread.idealThreadCount())
            env.insert("OMP_NUM_THREADS", str(num_cpu_pref))
            pastix_prec = "1" if self.fea.solver.PastixMixedPrecision else "0"
            env.insert("PASTIX_MIXED_PRECISION", pastix_prec)
            self.Calculix.setProcessEnvironment(env)

            self.cwd = QtCore.QDir.currentPath()
            fi = QtCore.QFileInfo(self.fea.inp_file_name)
            QtCore.QDir.setCurrent(fi.path())
            self.Calculix.start(self.fea.ccx_binary, ["-i", fi.baseName()])

            QApplication.restoreOverrideCursor()

    def stopCalculix(self):
        if self.Calculix.state() == QtCore.QProcess.ProcessState.Running:
            self.Calculix.kill()

    def select_analysis_type(self, analysis_type):
        if self.fea.solver.AnalysisType != analysis_type:
            self.fea.solver.AnalysisType = analysis_type
            self.form.pb_edit_inp.setEnabled(False)
            self.form.pb_run_ccx.setEnabled(False)

    def select_static_analysis(self):
        self.select_analysis_type("static")

    def select_frequency_analysis(self):
        self.select_analysis_type("frequency")

    def select_thermomech_analysis(self):
        self.select_analysis_type("thermomech")

    def select_check_mesh(self):
        self.select_analysis_type("check")

    def select_buckling_analysis(self):
        self.select_analysis_type("buckling")



# ***************************************************************************
# *   Copyright (c) 2017 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint flow velocity document object"
__author__ = "Markus Hovorka, Bernd Hahnebach, Uwe Stöhr"
__url__ = "https://www.freecad.org"

## @package constraint_flowvelocity
#  \ingroup FEM
#  \brief constraint flow velocity object

from . import base_fempythonobject


class ConstraintFlowVelocity(base_fempythonobject.BaseFemPythonObject):

    Type = "Fem::ConstraintFlowVelocity"

    def __init__(self, obj):
        super().__init__(obj)
        obj.addProperty(
            "App::PropertyVelocity", "VelocityX", "Parameter", "Velocity in X-direction"
        )
        obj.setPropertyStatus("VelocityX", "LockDynamic")
        obj.addProperty(
            "App::PropertyString",
            "VelocityXFormula",
            "Parameter",
            "Velocity formula in X-direction",
        )
        obj.setPropertyStatus("VelocityXFormula", "LockDynamic")
        obj.addProperty(
            "App::PropertyBool", "VelocityXUnspecified", "Parameter", "Use velocity in X-direction"
        )
        obj.setPropertyStatus("VelocityXUnspecified", "LockDynamic")
        obj.VelocityXUnspecified = True
        obj.addProperty(
            "App::PropertyBool",
            "VelocityXHasFormula",
            "Parameter",
            "Use formula for velocity in X-direction",
        )
        obj.setPropertyStatus("VelocityXHasFormula", "LockDynamic")

        obj.addProperty(
            "App::PropertyVelocity", "VelocityY", "Parameter", "Velocity in Y-direction"
        )
        obj.setPropertyStatus("VelocityY", "LockDynamic")
        obj.addProperty(
            "App::PropertyString",
            "VelocityYFormula",
            "Parameter",
            "Velocity formula in Y-direction",
        )
        obj.setPropertyStatus("VelocityYFormula", "LockDynamic")
        obj.addProperty(
            "App::PropertyBool", "VelocityYUnspecified", "Parameter", "Use velocity in Y-direction"
        )
        obj.setPropertyStatus("VelocityYUnspecified", "LockDynamic")
        obj.VelocityYUnspecified = True
        obj.addProperty(
            "App::PropertyBool",
            "VelocityYHasFormula",
            "Parameter",
            "Use formula for velocity in Y-direction",
        )
        obj.setPropertyStatus("VelocityYHasFormula", "LockDynamic")

        obj.addProperty(
            "App::PropertyVelocity", "VelocityZ", "Parameter", "Velocity in z-direction"
        )
        obj.setPropertyStatus("VelocityZ", "LockDynamic")
        obj.addProperty(
            "App::PropertyString",
            "VelocityZFormula",
            "Parameter",
            "Velocity formula in Z-direction",
        )
        obj.setPropertyStatus("VelocityZFormula", "LockDynamic")
        obj.addProperty(
            "App::PropertyBool", "VelocityZUnspecified", "Parameter", "Use velocity in Z-direction"
        )
        obj.setPropertyStatus("VelocityZUnspecified", "LockDynamic")
        obj.VelocityZUnspecified = True
        obj.addProperty(
            "App::PropertyBool",
            "VelocityZHasFormula",
            "Parameter",
            "Use formula for velocity in Z-direction",
        )
        obj.setPropertyStatus("VelocityZHasFormula", "LockDynamic")

        obj.addProperty(
            "App::PropertyBool", "NormalToBoundary", "Parameter", "Flow is in normal direction"
        )
        obj.setPropertyStatus("NormalToBoundary", "LockDynamic")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path.Base.Generator.threadmilling as threadmilling
import math

from CAMTests.PathTestUtils import PathTestBase


def radii(internal, major, minor, toolDia, toolCrest):
    """test radii function for simple testing"""
    return (minor, major)


class TestPathThreadMillingGenerator(PathTestBase):
    """Test thread milling generator."""

    def test00(self):
        """Verify thread commands for a single thread"""

        center = FreeCAD.Vector()
        cmd = "G2"
        zStart = 0
        zFinal = 1
        pitch = 1
        radius = 3
        leadInOut = False
        elevator = 2

        path, start = threadmilling.generate(
            center, cmd, zStart, zFinal, pitch, radius, leadInOut, elevator, None
        )

        gcode = [
            "G0 X0.000000 Y2.000000",
            "G0 Z0.000000",
            "G1 Y3.000000",
            "G2 J-3.000000 Y-3.000000 Z0.500000",
            "G2 J3.000000 Y3.000000 Z1.000000",
            "G1 X0.000000 Y2.000000",
        ]
        self.assertEqual([p.toGCode() for p in path], gcode)
        self.assertCoincide(start, FreeCAD.Vector(0, 2, zFinal))

    def test01(self):
        """Verify thread commands for a thwo threads"""

        center = FreeCAD.Vector()
        cmd = "G2"
        zStart = 0
        zFinal = 2
        pitch = 1
        radius = 3
        leadInOut = False
        elevator = 2

        path, start = threadmilling.generate(
            center, cmd, zStart, zFinal, pitch, radius, leadInOut, elevator, None
        )

        gcode = [
            "G0 X0.000000 Y2.000000",
            "G0 Z0.000000",
            "G1 Y3.000000",
            "G2 J-3.000000 Y-3.000000 Z0.500000",
            "G2 J3.000000 Y3.000000 Z1.000000",
            "G2 J-3.000000 Y-3.000000 Z1.500000",
            "G2 J3.000000 Y3.000000 Z2.000000",
            "G1 X0.000000 Y2.000000",
        ]
        self.assertEqual([p.toGCode() for p in path], gcode)
        self.assertCoincide(start, FreeCAD.Vector(0, 2, zFinal))

    def test02(self):
        """Verify thread commands for a one and a half threads"""

        center = FreeCAD.Vector()
        cmd = "G2"
        zStart = 0
        zFinal = 1.5
        pitch = 1
        radius = 3
        leadInOut = False
        elevator = 2

        path, start = threadmilling.generate(
            center, cmd, zStart, zFinal, pitch, radius, leadInOut, elevator, None
        )

        gcode = [
            "G0 X0.000000 Y2.000000",
            "G0 Z0.000000",
            "G1 Y3.000000",
            "G2 J-3.000000 Y-3.000000 Z0.500000",
            "G2 J3.000000 Y3.000000 Z1.000000",
            "G2 J-3.000000 Y-3.000000 Z1.500000",
            "G1 X0.000000 Y-2.000000",
        ]
        self.assertEqual([p.toGCode() for p in path], gcode)
        self.assertCoincide(start, FreeCAD.Vector(0, -2, zFinal))

    def test03(self):
        """Verify thread commands for a one and 3 quarter threads"""

        center = FreeCAD.Vector()
        cmd = "G2"
        zStart = 0
        zFinal = 1.75
        pitch = 1
        radius = 3
        leadInOut = False
        elevator = 2

        path, start = threadmilling.generate(
            center, cmd, zStart, zFinal, pitch, radius, leadInOut, elevator, None
        )

        gcode = [
            "G0 X0.000000 Y2.000000",
            "G0 Z0.000000",
            "G1 Y3.000000",
            "G2 J-3.000000 Y-3.000000 Z0.500000",
            "G2 J3.000000 Y3.000000 Z1.000000",
            "G2 J-3.000000 Y-3.000000 Z1.500000",
            #'(------- finish-thread -------)',
            "G2 J3.000000 X-3.000000 Y0.000000 Z1.750000",
            #'(------- finish-thread -------)',
            "G1 X-2.000000 Y0.000000",
        ]
        self.assertEqual([p.toGCode() for p in path], gcode)
        self.assertCoincide(start, FreeCAD.Vector(-2, 0, zFinal))

    def test04(self):
        """Verify thread commands for a one and 3 quarter threads - CCW"""

        center = FreeCAD.Vector()
        cmd = "G3"
        zStart = 0
        zFinal = 1.75
        pitch = 1
        radius = 3
        leadInOut = False
        elevator = 2

        path, start = threadmilling.generate(
            center, cmd, zStart, zFinal, pitch, radius, leadInOut, elevator, None
        )

        gcode = [
            "G0 X0.000000 Y2.000000",
            "G0 Z0.000000",
            "G1 Y3.000000",
            "G3 J-3.000000 Y-3.000000 Z0.500000",
            "G3 J3.000000 Y3.000000 Z1.000000",
            "G3 J-3.000000 Y-3.000000 Z1.500000",
            #'(------- finish-thread -------)',
            "G3 J3.000000 X3.000000 Y0.000000 Z1.750000",
            #'(------- finish-thread -------)',
            "G1 X2.000000 Y0.000000",
        ]
        self.assertEqual([p.toGCode() for p in path], gcode)
        self.assertCoincide(start, FreeCAD.Vector(2, 0, zFinal))

    def test10(self):
        """Verify lead in/out commands for a single thread"""

        center = FreeCAD.Vector()
        cmd = "G2"
        zStart = 0
        zFinal = 1
        pitch = 1
        radius = 3
        leadInOut = True
        elevator = 2

        path, start = threadmilling.generate(
            center, cmd, zStart, zFinal, pitch, radius, leadInOut, elevator, None
        )

        gcode = [
            "G0 X0.000000 Y2.000000",
            "G0 Z0.000000",
            #'(------- lead-in -------)',
            "G2 J0.500000 Y3.000000",
            #'(------- lead-in -------)',
            "G2 J-3.000000 Y-3.000000 Z0.500000",
            "G2 J3.000000 Y3.000000 Z1.000000",
            #'(------- lead-out -------)',
            "G2 I0.000000 J-0.500000 X0.000000 Y2.000000",
            #'(------- lead-out -------)',
        ]
        self.assertEqual([p.toGCode() for p in path], gcode)
        self.assertCoincide(start, FreeCAD.Vector(0, 2, zFinal))

    def test11(self):
        """Verify lead in/out commands for one and a half threads"""

        center = FreeCAD.Vector()
        cmd = "G2"
        zStart = 0
        zFinal = 1.5
        pitch = 1
        radius = 3
        leadInOut = True
        elevator = 2

        path, start = threadmilling.generate(
            center, cmd, zStart, zFinal, pitch, radius, leadInOut, elevator, None
        )

        gcode = [
            "G0 X0.000000 Y2.000000",
            "G0 Z0.000000",
            #'(------- lead-in -------)',
            "G2 J0.500000 Y3.000000",
            #'(------- lead-in -------)',
            "G2 J-3.000000 Y-3.000000 Z0.500000",
            "G2 J3.000000 Y3.000000 Z1.000000",
            "G2 J-3.000000 Y-3.000000 Z1.500000",
            #'(------- lead-out -------)',
            "G2 I0.000000 J0.500000 X0.000000 Y-2.000000",
            #'(------- lead-out -------)',
        ]
        self.assertEqual([p.toGCode() for p in path], gcode)
        self.assertCoincide(start, FreeCAD.Vector(0, -2, zFinal))



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2025 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

import json
import os
import threading
import datetime
from pathlib import Path
from typing import Dict

import addonmanager_freecad_interface as fci
from Addon import Addon
from AddonCatalog import AddonCatalog


def most_recent_update(directory_path: str) -> datetime.datetime:
    """Walk through a path and return the most recent update time."""
    path = Path(directory_path)

    if not path.exists():
        raise FileNotFoundError(f"Directory not found: {path}")
    if not path.is_dir():
        raise NotADirectoryError(f"Path is not a directory: {path}")

    latest_time = datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)
    for file_path in path.rglob("*"):
        if file_path.is_file():
            try:
                mtime = datetime.datetime.fromtimestamp(file_path.stat().st_mtime).astimezone()
                if mtime > latest_time:
                    latest_time = mtime
            except (PermissionError, OSError):
                continue
    return latest_time


class InstallationManifest:
    """The installation manifest tracks installation and updates of addons so that the Addon Manager
    knows what branch is currently installed (this gets reflected in the change-branch UI for addons
    that list multiple branches in the catalog). It also tracks the original installation time and
    last updated time, which may eventually get used in the interface."""

    lock = threading.Lock()

    path_to_manifest_file = ""

    def __init__(self, catalog: AddonCatalog = None):
        self._manifest = {}
        self.old_backups = []
        self.unrecognized_directories = []
        if not InstallationManifest.path_to_manifest_file:
            InstallationManifest.path_to_manifest_file = os.path.join(
                fci.DataPaths().mod_dir, "manifest.json"
            )
        if not os.path.exists(InstallationManifest.path_to_manifest_file):
            self._migrate_to_manifest_file(catalog)
            self.write_manifest()
        else:
            self.load_manifest()
        if catalog:
            self._scan_mod_path_for_extras(catalog)

    def _migrate_to_manifest_file(self, catalog: AddonCatalog):
        if not os.path.exists(fci.DataPaths().mod_dir):
            os.makedirs(fci.DataPaths().mod_dir)
        dirs_in_mod = os.listdir(fci.DataPaths().mod_dir)
        for addon_id in dirs_in_mod:
            if Path(os.path.join(fci.DataPaths().mod_dir, addon_id)).is_dir():
                branches = []
                if catalog:
                    branches = catalog.get_available_branches(addon_id)
                if branches:
                    branch_display_name = branches[0]
                    self._manifest[addon_id] = {
                        "addon_id": addon_id,
                        "migrated": True,
                        "first_installed": datetime.datetime.fromtimestamp(
                            0, tz=datetime.timezone.utc
                        ).isoformat(),
                        "last_updated": most_recent_update(
                            os.path.join(fci.DataPaths().mod_dir, addon_id)
                        ).isoformat(),
                        "branch_display_name": branch_display_name,
                        "extra_files": [],
                        "freecad_version": "",
                    }
            else:
                fci.Console.PrintMessage("Migrate to Manifest, skipping file: " + addon_id + "\n")

    def load_manifest(self):
        """Load the manifest from the disk"""
        with InstallationManifest.lock:
            with open(self.path_to_manifest_file, "r") as f:
                self._manifest = json.load(f)

    def write_manifest(self):
        """Write the manifest to the disk"""
        with InstallationManifest.lock:
            with open(self.path_to_manifest_file, "w") as f:
                json.dump(self._manifest, f, indent=2)

    def _scan_mod_path_for_extras(self, catalog: AddonCatalog):
        dirs_in_mod = os.listdir(fci.DataPaths().mod_dir)
        for addon_id in dirs_in_mod:
            if not os.path.isdir(os.path.join(fci.DataPaths().mod_dir, addon_id)):
                continue
            branches = catalog.get_available_branches(addon_id)
            if not branches:
                if "backup" in addon_id:
                    fci.Console.PrintMessage("Found old backup directory: " + addon_id + "\n")
                    self.old_backups.append(addon_id)
                else:
                    fci.Console.PrintMessage(
                        "Found addon not in main AM catalog: " + addon_id + "\n"
                    )
                    self.unrecognized_directories.append(addon_id)

    def record_new_installation(self, addon_id: str, addon: Addon, extra_files: list = None):
        """Record the first installation of an addon.
        :param addon_id: The addon ID
        :param addon: The Addon object
        :param extra_files: A list of extra files to record as installed
        """
        self._manifest[addon_id] = {
            "addon_id": addon_id,
            "migrated": False,
            "first_installed": datetime.datetime.now(datetime.timezone.utc).isoformat(),
            "last_updated": datetime.datetime.now(datetime.timezone.utc).isoformat(),
            "branch_display_name": addon.branch_display_name,
            "extra_files": [] if extra_files is None else extra_files,
            "freecad_version": fci.Version()[0:3],
        }
        self.write_manifest()

    def record_update(self, addon_id: str, addon: Addon, extra_files: list = None):
        """Record the update of an addon that was already installed. Will raise an exception if the
        addon is not already in the manifest.
        :param addon_id: The addon ID
        :param addon: The Addon object
        :param extra_files: A list of extra files to record as installed (typically copied FCmacro files)
        """
        self._manifest[addon_id]["last_updated"] = datetime.datetime.now(
            datetime.timezone.utc
        ).isoformat()
        self._manifest[addon_id]["branch_display_name"] = addon.branch_display_name
        self._manifest[addon_id]["freecad_version"] = fci.Version()[0:3]
        if extra_files:
            all_extra_files = set(self._manifest[addon_id]["extra_files"])
            all_extra_files.update(extra_files)
            self._manifest[addon_id]["extra_files"] = list(all_extra_files)
        self.write_manifest()

    def remove(self, addon_id: str):
        """Remove an addon from the manifest (that is, it was uninstalled -- we don't record the
        uninstallation, it just gets dropped)."""
        self._manifest.pop(addon_id, None)
        self.write_manifest()

    def contains(self, addon_id: str) -> bool:
        return addon_id in self._manifest

    def get_addon_info(self, addon_id: str) -> Dict:
        return self._manifest[addon_id]
